from .base import TestBase
from ..loader import (CommunityLauncher, IPv8CommunityLoader, after, bootstrapper, kwargs, name, overlay, precondition,
                      set_in_session, walk_strategy)


class MockCommunity:

    def __init__(self, peer, endpoint, network, *args, **kw_args):
        self.peer = peer
        self.endpoint = endpoint
        self.network = network
        self.args = args
        self.kwargs = kw_args
        self.bootstrappers = []


class MockWalk:

    def __init__(self, community, some_attribute):
        self.overlay = community
        self.some_attribute = some_attribute


class MockWalk2:
    pass


class MockBootstrapper:

    def __init__(self, some_attribute):
        self.some_attribute = some_attribute


class MockBootstrapper2:
    pass


class MockSession:

    launch_condition1 = True
    launch_condition2 = False
    some_attribute1 = "I am a string :)"
    some_attribute2 = 1337
    community = None


class MockOverlayProvider:
    endpoint = None
    network = None

    def __init__(self):
        self.overlays = []
        self.strategies = []


class StagedCommunityLauncher(CommunityLauncher):

    def not_before(self):
        return ['CommunityLauncher1', 'CommunityLauncher2']

    def should_launch(self, session: MockSession):
        return session.launch_condition1 and not session.launch_condition2

    def get_overlay_class(self):
        return MockCommunity

    def get_kwargs(self, session: MockSession):
        return {
            'kw1': session.some_attribute1,
            'kw2': session.some_attribute2
        }

    def get_walk_strategies(self):
        return [(MockWalk, {'some_attribute': 4}, 20)]

    def get_bootstrappers(self, session):
        return [(MockBootstrapper, {'some_attribute': 4})]

    def finalize(self, ipv8, session: MockSession, community: MockCommunity):
        session.community = community
        return super()


class TestCommunityLauncher(TestBase):

    def setUp(self):
        self.staged_launcher = StagedCommunityLauncher()
        return super()

    def test_not_before_list(self):
        """
        Check that the not_before decorator with multiple arguments equals the not_before() definition.
        """
        @after('CommunityLauncher1', 'CommunityLauncher2')
        class DecoratedCommunityLauncher(CommunityLauncher):
            pass

        self.assertListEqual(self.staged_launcher.not_before(), DecoratedCommunityLauncher().not_before())

    def test_not_before_multiple(self):
        """
        Check that multiple not_before decorators with an argument equals the not_before() definition.
        """
        @after('CommunityLauncher2')
        @after('CommunityLauncher1')
        class DecoratedCommunityLauncher(CommunityLauncher):
            pass

        self.assertListEqual(self.staged_launcher.not_before(), DecoratedCommunityLauncher().not_before())

    def test_should_launch_single(self):
        """
        Check that a validated single launch condition causes should_launch() to return True.
        """
        @precondition('session.launch_condition1')
        class DecoratedCommunityLauncher(CommunityLauncher):
            pass

        self.assertTrue(DecoratedCommunityLauncher().should_launch(MockSession()))

    def test_shouldnt_launch_single(self):
        """
        Check that an invalid single launch condition causes should_launch() to return False.
        """
        @precondition('session.launch_condition2')
        class DecoratedCommunityLauncher(CommunityLauncher):
            pass

        self.assertFalse(DecoratedCommunityLauncher().should_launch(MockSession()))

    def test_should_launch_multiple(self):
        """
        Check that a validated multiple launch conditions causes should_launch() to return True.
        """
        @precondition('session.launch_condition1')
        @precondition('not session.launch_condition2')
        class DecoratedCommunityLauncher(CommunityLauncher):
            pass

        self.assertTrue(DecoratedCommunityLauncher().should_launch(MockSession()))
        self.assertTrue(self.staged_launcher.should_launch(MockSession()))

    def test_shouldnt_launch_multiple(self):
        """
        Check that an invalid condition for multiple launch conditions causes should_launch() to return False.
        """
        @precondition('session.launch_condition1')
        @precondition('session.launch_condition2')
        class DecoratedCommunityLauncher(CommunityLauncher):
            pass

        self.assertFalse(DecoratedCommunityLauncher().should_launch(MockSession()))

    def test_overlay_class_from_str(self):
        """
        Check if a Community string specification can be lazy-loaded through the overlay_class decorator.
        """
        @overlay(self.__class__.__module__, 'MockCommunity')
        class DecoratedCommunityLauncher(CommunityLauncher):
            pass

        self.assertEqual(self.staged_launcher.get_overlay_class(), DecoratedCommunityLauncher().get_overlay_class())
        self.assertSetEqual({self.__class__.__module__}, DecoratedCommunityLauncher.hiddenimports)

    def test_overlay_class_from_class(self):
        """
        Check if a Community class can be lazy-loaded through the overlay_class decorator.
        """
        @overlay(MockCommunity)
        class DecoratedCommunityLauncher(CommunityLauncher):
            pass

        self.assertEqual(self.staged_launcher.get_overlay_class(), DecoratedCommunityLauncher().get_overlay_class())

    def test_overlay_class_from_function(self):
        """
        Check if a Community class (functional representation) can be lazy-loaded
        through the overlay_class decorator.
        """

        def MockCommunityFunction():
            return MockCommunity

        @overlay(MockCommunityFunction)
        class DecoratedCommunityLauncher(CommunityLauncher):
            pass

        self.assertEqual(self.staged_launcher.get_overlay_class(),
                         DecoratedCommunityLauncher().get_overlay_class())

    def test_walk_strategy_from_str(self):
        """
        Check if adding a walk strategy string specification is successful.
        """
        @walk_strategy(self.__class__.__module__, 'MockWalk', kw_args={'some_attribute': 4})
        class DecoratedCommunityLauncher(CommunityLauncher):
            pass

        self.assertListEqual(self.staged_launcher.get_walk_strategies(),
                             DecoratedCommunityLauncher().get_walk_strategies())
        self.assertSetEqual({self.__class__.__module__}, DecoratedCommunityLauncher.hiddenimports)

    def test_walk_strategy_from_class(self):
        """
        Check if adding a walk strategy from a DiscoveryStrategy class is successful.
        """
        @walk_strategy(MockWalk, kw_args={'some_attribute': 4})
        class DecoratedCommunityLauncher(CommunityLauncher):
            pass

        self.assertListEqual(self.staged_launcher.get_walk_strategies(),
                             DecoratedCommunityLauncher().get_walk_strategies())

    def test_walk_strategy_from_function(self):
        """
        Check if adding a walk strategy from a function is successful.
        """

        def MockWalkFunction():
            return MockWalk

        @walk_strategy(MockWalkFunction, kw_args={'some_attribute': 4})
        class DecoratedCommunityLauncher(CommunityLauncher):
            pass

        self.assertListEqual(self.staged_launcher.get_walk_strategies(),
                             DecoratedCommunityLauncher().get_walk_strategies())

    def test_walk_strategy_multiple(self):
        """
        Check if adding multiple walk strategies is successful.
        """
        @walk_strategy(MockWalk2, target_peers=-1)
        @walk_strategy(self.__class__.__module__, 'MockWalk', kw_args={'some_attribute': 4})
        class DecoratedCommunityLauncher(CommunityLauncher):
            pass

        self.assertListEqual([(MockWalk, {'some_attribute': 4}, 20), (MockWalk2, {}, -1)],
                             DecoratedCommunityLauncher().get_walk_strategies())
        self.assertSetEqual({self.__class__.__module__}, DecoratedCommunityLauncher.hiddenimports)

    def test_bootstrapper_from_str(self):
        """
        Check if adding a bootstrapper string specification is successful.
        """
        @bootstrapper(self.__class__.__module__, 'MockBootstrapper', kw_args={'some_attribute': 4})
        class DecoratedCommunityLauncher(CommunityLauncher):
            pass

        self.assertListEqual(self.staged_launcher.get_bootstrappers(MockSession()),
                             DecoratedCommunityLauncher().get_bootstrappers(MockSession()))
        self.assertSetEqual({self.__class__.__module__}, DecoratedCommunityLauncher.hiddenimports)

    def test_bootstrapper_from_class(self):
        """
        Check if adding a bootstrapper from a Bootstrapper class is successful.
        """
        @bootstrapper(MockBootstrapper, kw_args={'some_attribute': 4})
        class DecoratedCommunityLauncher(CommunityLauncher):
            pass

        self.assertListEqual(self.staged_launcher.get_bootstrappers(MockSession()),
                             DecoratedCommunityLauncher().get_bootstrappers(MockSession()))

    def test_bootstrapper_from_function(self):
        """
        Check if adding a bootstrapper from a function is successful.
        """

        def mock_bootstrapper():
            return MockBootstrapper

        @bootstrapper(mock_bootstrapper, kw_args={'some_attribute': 4})
        class DecoratedCommunityLauncher(CommunityLauncher):
            pass

        self.assertListEqual(self.staged_launcher.get_bootstrappers(MockSession()),
                             DecoratedCommunityLauncher().get_bootstrappers(MockSession()))

    def test_bootstrapper_multiple(self):
        """
        Check if adding multiple bootstrappers is successful.
        """
        @bootstrapper(MockBootstrapper2)
        @bootstrapper(self.__class__.__module__, 'MockBootstrapper', kw_args={'some_attribute': 4})
        class DecoratedCommunityLauncher(CommunityLauncher):
            pass

        self.assertListEqual([(MockBootstrapper, {'some_attribute': 4}), (MockBootstrapper2, {})],
                             DecoratedCommunityLauncher().get_bootstrappers(MockSession()))
        self.assertSetEqual({self.__class__.__module__}, DecoratedCommunityLauncher.hiddenimports)

    def test_set_in_session(self):
        """
        Check if set_in_session correctly sets the attribute of the session.
        """
        @set_in_session('community')
        class DecoratedCommunityLauncher(CommunityLauncher):
            pass

        session = MockSession()
        community = MockCommunity(None, None, None)
        DecoratedCommunityLauncher().finalize(None, session, community)

        self.assertEqual(community, session.community)

    def test_kwargs(self):
        """
        Check if the kwargs decorator correctly passes keyword arguments.
        """
        @kwargs(kw1='session.some_attribute1', kw2='session.some_attribute2')
        class DecoratedCommunityLauncher(CommunityLauncher):
            pass

        session = MockSession()
        self.assertDictEqual(self.staged_launcher.get_kwargs(session), DecoratedCommunityLauncher().get_kwargs(session))

    def test_name(self):
        """
        Check if the name of a launcher can be set, using the name decorator.
        """
        @name('Some Name')
        class DecoratedCommunityLauncher(CommunityLauncher):
            pass

        self.assertEqual('Some Name', DecoratedCommunityLauncher().get_name())

    def test_no_name(self):
        """
        Check if the name of a launcher is equal to the Community class name by default.
        """
        @overlay(MockCommunity)
        class DecoratedCommunityLauncher(CommunityLauncher):
            pass

        self.assertEqual('DecoratedCommunityLauncher', DecoratedCommunityLauncher().get_name())


class TestCommunityLoader(TestBase):

    def setUp(self):
        self.community_loader = IPv8CommunityLoader()
        self.overlay_provider = MockOverlayProvider()
        self.session = MockSession()

    def test_load_community(self):
        """
        Check if a CommunityLauncher is launched correctly.
        """
        self.community_loader.set_launcher(StagedCommunityLauncher())

        self.community_loader.load(self.overlay_provider, self.session)

        self.assertEqual(1, len(self.overlay_provider.overlays))
        self.assertEqual(1, len(self.overlay_provider.strategies))

        loaded_overlay = self.overlay_provider.overlays[0]
        loaded_strategy = self.overlay_provider.strategies[0]

        self.assertIsInstance(loaded_overlay, MockCommunity)
        self.assertIsInstance(loaded_strategy[0], MockWalk)
        self.assertEqual(loaded_overlay, loaded_strategy[0].overlay)
        self.assertEqual(4, loaded_strategy[0].some_attribute)
        self.assertEqual(20, loaded_strategy[1])

    def test_protect_infinite_loop(self):
        """
        Check if the CommunityLoader raises an error when it encounters a circular dependency.
        """
        @name("A")
        @overlay(MockCommunity)
        @after("B")
        class A(CommunityLauncher):
            pass

        @name("B")
        @overlay(MockCommunity)
        @after("A")
        class B(CommunityLauncher):
            pass

        self.community_loader.set_launcher(A())
        self.community_loader.set_launcher(B())

        self.assertRaises(RuntimeError, self.community_loader.load, self.overlay_provider, self.session)



from __future__ import division

import os
from functools import reduce


def secure_randint(bitspace):
    delbits = 8 - (bitspace % 8)
    bytez = os.urandom(bitspace // 8)
    if delbits > 0:
        b = (os.urandom(1)[0] & (0xFF >> delbits)).to_bytes(1, 'big', signed=False)
        bytez = b + bytez
    return reduce(lambda a, b: (a << 8) + b, bytez, 0)



import time

from ..base import TestBase
from ..mocking.community import MockCommunity
from ..mocking.endpoint import MockEndpointListener
from ...peerdiscovery.churn import RandomChurn


class TestChurn(TestBase):

    def setUp(self):
        super(TestChurn, self).setUp()

        node_count = 2
        self.overlays = [MockCommunity() for _ in range(node_count)]
        self.strategies = [RandomChurn(self.overlays[i], ping_interval=0.0) for i in range(node_count)]

    async def tearDown(self):
        for overlay in self.overlays:
            await overlay.unload()
        return await super(TestChurn, self).tearDown()

    async def test_keep_reachable(self):
        """
        Check if we don't remove reachable nodes.
        """
        peer = self.overlays[1].my_peer
        fake_last_response = time.time() - 30
        peer.last_response = fake_last_response
        self.overlays[0].network.add_verified_peer(peer)

        # Node inactive! We should ping it!
        self.strategies[0].take_step()

        await self.deliver_messages()

        # Node should have responded by now and not be removed.
        self.assertNotEqual(peer.last_response, fake_last_response)
        self.strategies[0].take_step()

        self.assertEqual(len(self.overlays[0].network.verified_peers), 1)

    async def test_remove_unreachable(self):
        """
        Check if we remove unreachable nodes.
        """
        peer = self.overlays[1].my_peer
        fake_last_response = time.time() - 30
        peer.last_response = fake_last_response
        self.overlays[0].network.add_verified_peer(peer)
        self.overlays[1].endpoint.close()

        # Node inactive! We should ping it!
        self.strategies[0].take_step()

        await self.deliver_messages()

        # Node should not have responded.
        self.assertEqual(peer.last_response, fake_last_response)
        # Nothing will happen the next 30 seconds, so we forward the clock a bit.
        peer.last_response = 1
        # Now the node should be removed.
        self.strategies[0].take_step()

        self.assertEqual(len(self.overlays[0].network.verified_peers), 0)

    async def test_no_nodes(self):
        """
        Nothing should happen if we have no nodes to check.
        """
        self.strategies[0].sample_size = 0  # Don't check anything
        peer = self.overlays[1].my_peer
        peer.last_response = 1  # 50 years ago
        self.overlays[0].network.add_verified_peer(peer)

        self.strategies[0].take_step()

        await self.deliver_messages()

        self.assertEqual(len(self.overlays[0].network.verified_peers), 1)

    async def test_ping_timeout(self):
        """
        Don't overload inactive nodes with pings, send it once within some timeout.
        """
        peer = self.overlays[1].my_peer
        peer.last_response = time.time() - 30
        self.overlays[0].network.add_verified_peer(peer)
        self.strategies[0].ping_interval = 10000.0
        # Hook up listener
        sniffer = MockEndpointListener(self.overlays[1].endpoint)

        # Node inactive! We should ping it!
        self.strategies[0].take_step()
        # Only once though
        self.strategies[0].take_step()

        await self.deliver_messages()

        self.assertEqual(len(sniffer.received_packets), 1)

    async def test_ping_timeout_resend(self):
        """
        Don't overload inactive nodes with pings, send it again after some timeout.
        """
        peer = self.overlays[1].my_peer
        peer.last_response = time.time() - 30
        self.overlays[0].network.add_verified_peer(peer)
        self.strategies[0].ping_interval = -1.0
        # Hook up listener
        sniffer = MockEndpointListener(self.overlays[1].endpoint)

        # Node inactive! We should ping it!
        self.strategies[0].take_step()
        # Timeout has passed, send it again.
        self.strategies[0].take_step()

        await self.deliver_messages()

        self.assertEqual(len(sniffer.received_packets), 2)



import logging
from asyncio import Future
from random import choice
from socket import error, gethostbyname
from threading import Thread
from time import time
from typing import Iterable

from ..bootstrapper_interface import Bootstrapper
from ...community import Community
from ...messaging.interfaces.udp.endpoint import UDPv4Address
from ...types import Address
from ...util import succeed


class DispersyBootstrapper(Bootstrapper):

    def __init__(self, ip_addresses, dns_addresses, bootstrap_timeout=30.0):
        self.ip_addresses = [UDPv4Address(*addr) for addr in ip_addresses]
        self.dns_addresses = dns_addresses

        self.bootstrap_timeout = bootstrap_timeout

        self.last_bootstrap = 0

    def resolve_dns_bootstrap_addresses(self):
        """
        Resolve the bootstrap server DNS names defined in ``dns_addresses`` and insert them into ``ip_addresses``.
        """
        def resolve_addresses(dns_names, ip_addresses):
            current_addresses = ip_addresses[:]  # Copy the existing addresses (don't loop through our additions)
            for (address, port) in dns_names:
                try:
                    resolved_address = UDPv4Address(gethostbyname(address), port)
                    if resolved_address not in current_addresses:
                        # NOTE: append() is thread-safe. Don't call remove() here!
                        ip_addresses.append(resolved_address)
                except error:
                    logging.info("Unable to resolve bootstrap DNS address (%s, %d)", address, port)

        resolution_thread = Thread(name="resolve_dns_bootstrap_addresses",
                                   target=resolve_addresses,
                                   args=(self.dns_addresses, self.ip_addresses),
                                   daemon=True)
        resolution_thread.start()

    def initialize(self, overlay: Community) -> Future:
        overlay.network.blacklist.extend(self.ip_addresses)
        self.resolve_dns_bootstrap_addresses()
        return succeed(True)

    async def get_addresses(self, overlay: Community, timeout: float) -> Iterable[Address]:
        if time() - self.last_bootstrap < self.bootstrap_timeout:
            return []
        logging.debug("Bootstrapping %s, current peers %d", overlay.__class__.__name__, len(overlay.get_peers()))
        self.last_bootstrap = time()
        for socket_address in self.ip_addresses:
            overlay.ensure_blacklisted(socket_address)
            overlay.walk_to(socket_address)
        return []

    def keep_alive(self, overlay: Community) -> None:
        if self.ip_addresses:
            address = choice(self.ip_addresses)
            overlay.ensure_blacklisted(address)
            overlay.walk_to(address)

    def blacklist(self) -> Iterable[Address]:
        return self.ip_addresses

    def unload(self) -> None:
        pass



from ...base import TestBase
from ....attestation.tokentree.token import Token
from ....keyvault.crypto import ECCrypto


class TestToken(TestBase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.test_data = b"1234567890abcdefghijklmnopqrstuvwxyz" * 69
        cls.test_data_hash = b"[B\xc79lC\x07\xc88T\xe4yVN0+\x9e}\xc1\x1e\xfc\x88'm\x8d7\xe1\xa4*5\x06$"
        cls.test_public_key = ECCrypto().key_from_public_bin(b'LibNaCLPK:\xc8\xf38};U\xe4\xd5\xf7\xfd\xbc+J!\xbe\xba'
                                                             b'\x81M\xda\xef\xb7\x8c\xacL\x1eZ\x9d\xaf\xaaX+&\xac\xe2'
                                                             b'\xd2\xdd\x86\xa9\x97\xb8T\x9b\x82\xc1>\xa2\r\x11?\xef'
                                                             b'\x137\xf1\xdc!\x7f\x9fW\xe7\x11.\xe2\xc8)')
        cls.test_signature = (b'2\x05\xb3\xf24\xad\xc8)\xf5\x8bh+(0\x153\x01L\xfd-\x8e\x8bWr\xcbz\xd7\xbdEt\\\xd7\xe4'
                              b',\x80g\xbe\x7fT\x9c8\xb0\xc5\x80\xe4\xfa\xec\xad\xceY\xb8\xcek\x8bp\x81\xb0,j7`\x0f'
                              b'\xf8\x0f')

    def setUp(self) -> None:
        super(TestToken, self).setUp()

        self.private_key = ECCrypto().generate_key("curve25519")

    def test_create_private_token(self) -> None:
        """
        Check if a token is correctly created with a private key.
        """
        token = Token(b"test_previous_token_hash", content=self.test_data, private_key=self.private_key)

        self.assertEqual(b"test_previous_token_hash", token.previous_token_hash)
        self.assertEqual(self.test_data, token.content)
        self.assertEqual(self.test_data_hash, token.content_hash)
        self.assertTrue(token.verify(self.private_key.pub()))

    def test_verify_token_illegal(self) -> None:
        """
        Check if a token does not verify for a different public key.
        """
        token = Token(b"test_previous_token_hash", content=self.test_data, private_key=self.private_key)
        other_key = ECCrypto().generate_key("curve25519").pub()

        self.assertFalse(token.verify(other_key))

    def test_create_public_token(self) -> None:
        """
        Check if a token is correctly loaded with a public key.
        """
        token = Token(b"test_previous_token_hash", content_hash=self.test_data_hash, signature=self.test_signature)

        self.assertEqual(b"test_previous_token_hash", token.previous_token_hash)
        self.assertIsNone(token.content)
        self.assertEqual(self.test_data_hash, token.content_hash)
        self.assertTrue(token.verify(self.test_public_key))

    def test_update_public_token(self) -> None:
        """
        Check if a token with only a content hash accepts the correct content.
        """
        token = Token(b"test_previous_token_hash", content_hash=self.test_data_hash, signature=self.test_signature)

        return_value = token.receive_content(self.test_data)

        self.assertTrue(return_value)
        self.assertEqual(b"test_previous_token_hash", token.previous_token_hash)
        self.assertEqual(self.test_data, token.content)
        self.assertEqual(self.test_data_hash, token.content_hash)
        self.assertTrue(token.verify(self.test_public_key))

    def test_update_public_token_illegal(self) -> None:
        """
        Check if a token with only a content hash rejects invalid content.
        """
        token = Token(b"test_previous_token_hash", content_hash=self.test_data_hash, signature=self.test_signature)

        return_value = token.receive_content(b"some other data")

        self.assertFalse(return_value)
        self.assertEqual(b"test_previous_token_hash", token.previous_token_hash)
        self.assertIsNone(token.content)
        self.assertEqual(self.test_data_hash, token.content_hash)
        self.assertTrue(token.verify(self.test_public_key))



"""
Copyright (c), Privacy By Design Foundation
All rights reserved.

This source code has been ported from https://github.com/privacybydesign/irmago
The authors of this file are not -in any way- affiliated with the original authors or organizations.
"""
import binascii
import calendar
import datetime
import hashlib
import time

from .....util import int2byte

ExpiryFactor = 60 * 60 * 24 * 7
metadataLength = 1 + 3 + 2 + 2 + 16


class metadataField(object):

    def __init__(self, length, offset):
        self.length = length
        self.offset = offset


versionField = metadataField(1, 0)
signingDateField = metadataField(3, 1)
validityField = metadataField(2, 4)
keyCounterField = metadataField(2, 6)
credentialID = metadataField(16, 8)


def int_to_str(n):
    hexInt = hex(n).lstrip('0x').rstrip('L')
    if (len(hexInt) % 2) == 1:
        hexInt = '0' + hexInt
    return binascii.unhexlify(hexInt)


def shortToByte(x):
    return int_to_str(x)[-2:]


class MetadataAttribute(object):

    def __init__(self, version):
        self.Int = 0
        self.pk = None
        self.Conf = None

        self.setField(versionField, version)
        self.setSigningDate()
        self.setKeyCounter(0)
        self.setExpiryDate()

    def Bytes(self):
        bytez = int_to_str(self.Int)
        if len(bytez) < metadataLength:
            bytez += b'\x00' * (metadataLength - len(bytez))
        return bytez

    def setField(self, field, value):
        bytez_array = [int2byte(c) if isinstance(c, int) else c for c in self.Bytes()]
        startindex = field.length - len(value)
        for i in range(field.length):
            if i < startindex:
                bytez_array[i + field.offset] = b'\x00'
            else:
                bytez_array[i + field.offset] = value[i - startindex:i - startindex + 1]

        self.Int = int(binascii.hexlify(b''.join(bytez_array)), 16)

    def field(self, field):
        return self.Bytes()[field.offset:field.offset + field.length]

    def setSigningDate(self, timestamp=None):
        if timestamp:
            self.setField(signingDateField, shortToByte(timestamp))
        else:
            self.setField(signingDateField, shortToByte(int(time.time() / ExpiryFactor)))

    def setKeyCounter(self, i):
        self.setField(keyCounterField, shortToByte(i))

    def SigningDate(self):
        bytez_array = [int2byte(c) if isinstance(c, int) else c for c in self.field(signingDateField)]
        bytez_array = bytez_array[1:]
        timestamp = int(binascii.hexlify(b''.join(bytez_array)), 16) * ExpiryFactor
        return timestamp

    def setValidityDuration(self, weeks):
        self.setField(validityField, shortToByte(weeks))

    def setExpiryDate(self):
        expiry = datetime.datetime.now()
        month = expiry.month - 1 + 6
        year = expiry.year + month // 12
        month = month % 12 + 1
        day = min(expiry.day, calendar.monthrange(year, month)[1])
        expiry = time.mktime(datetime.date(year, month, day).timetuple())
        signing = self.SigningDate()
        self.setValidityDuration(int((expiry - signing) / ExpiryFactor))

    def setExpiryDateFromTimestamp(self, expiry):
        signing = self.SigningDate()
        self.setValidityDuration(int((expiry - signing) / ExpiryFactor))

    def setCredentialTypeIdentifier(self, the_id):
        bytez = hashlib.sha256(the_id).digest()
        self.setField(credentialID, bytez[:16])


def make_attribute_list(cr, attribute_order=None, validity_signing=None):
    """
    cr =
        {
            u'attributes': { ... "name": "value" ... },
            u'credential': u'pbdf.nijmegen.address',
            u'keyCounter': 0,
            u'validity': 1570123936
        }

    :param attribute_order: the order in which to handle the keys
    :type attribute_order: list
    """
    meta = MetadataAttribute(b'\x03')
    meta.setKeyCounter(cr[u'keyCounter'])
    meta.setCredentialTypeIdentifier(cr[u'credential'].encode('utf-8'))
    if validity_signing:
        meta.setValidityDuration(validity_signing[0])
        meta.setSigningDate(validity_signing[1])
    else:
        meta.setSigningDate()
        meta.setExpiryDateFromTimestamp(cr[u'validity'])
    signing_date = int(binascii.hexlify(meta.field(signingDateField)), 16)

    attrs = [meta.Int]
    attr_map = cr[u"attributes"]
    attribute_order = attribute_order or attr_map.keys()

    for k in attribute_order:
        if attr_map.get(k, None) is None:
            attrs.append(0)
        elif not attr_map[k]:
            attrs.append(1)
        else:
            encoded = attr_map[k].encode('utf-8')
            v = int(binascii.hexlify(b''.join(int2byte(c) if isinstance(c, int) else c for c in encoded)), 16)
            v <<= 1
            v += 1
            attrs.append(v)

    return attrs, signing_date



from ..primitives.structs import BonehPublicKey, ipack, iunpack
from ..primitives.value import FP2Value
from ...identity_formats import Attestation

__all__ = ['BitPairAttestation', 'BonehAttestation']


class BitPairAttestation(object):
    """
    An attestation of a single bitpair of a larger Attestation.
    """

    def __init__(self, a, b, complement):
        self.a = a
        self.b = b
        self.complement = complement

    def compress(self):
        return self.a * self.b * self.complement

    def serialize(self):
        return (ipack(self.a.a) + ipack(self.a.b) + ipack(self.b.a) + ipack(self.b.b)
                + ipack(self.complement.a) + ipack(self.complement.b))

    @classmethod
    def unserialize(cls, s, p):
        rem = s
        nums = []
        while rem and len(nums) < 6:
            unpacked, rem = iunpack(rem)
            nums.append(unpacked)
        inits = [FP2Value(p, nums[0], nums[1]),
                 FP2Value(p, nums[2], nums[3]),
                 FP2Value(p, nums[4], nums[5])]
        return cls(*inits)


class BonehAttestation(Attestation):
    """
    An attestation for a public key of a value consisting of multiple bitpairs.
    """

    def __init__(self, PK, bitpairs, id_format=None):
        super(BonehAttestation, self).__init__()
        self.bitpairs = bitpairs
        self.PK = PK
        self.id_format = id_format

    def serialize(self):
        out = b''
        out += self.PK.serialize()
        for bitpair in self.bitpairs:
            out += bitpair.serialize()
        return out

    def serialize_private(self, PK):
        return self.serialize()

    @classmethod
    def unserialize(cls, s, id_format=None):
        PK = BonehPublicKey.unserialize(s)
        bitpairs = []
        rem = s[len(PK.serialize()):]
        while rem:
            attest = BitPairAttestation.unserialize(rem, PK.p)
            bitpairs.append(attest)
            rem = rem[len(attest.serialize()):]
        return cls(PK, bitpairs, id_format)

    @classmethod
    def unserialize_private(cls, SK, s, id_format=None):
        return cls.unserialize(s, id_format)



import hashlib
import os
import time
from asyncio import FIRST_COMPLETED, Future, gather, wait
from binascii import hexlify, unhexlify
from collections import defaultdict, deque
from itertools import zip_longest

from . import DHTError
from .churn import PingChurn
from .payload import (FindRequestPayload, FindResponsePayload, NodePacker, PingRequestPayload, PingResponsePayload,
                      SignedStrPayload, StoreRequestPayload, StoreResponsePayload, StrPayload)
from .routing import Node, RoutingTable, calc_node_id, distance
from .storage import Storage
from ..community import Community
from ..lazy_community import lazy_wrapper, lazy_wrapper_wd
from ..messaging.interfaces.udp.endpoint import UDPv4Address, UDPv6Address
from ..messaging.serialization import ListOf
from ..peerdiscovery.network import Network
from ..requestcache import RandomNumberCache, RequestCache
from ..taskmanager import task

PING_INTERVAL = 25

# Maximum number of seconds a token can remain valid
TOKEN_EXPIRATION_TIME = 600

DHT_ENTRY_STR = 0
DHT_ENTRY_STR_SIGNED = 1

MAX_ENTRY_SIZE = 170
MAX_ENTRY_AGE = 3600

# Maximum number of nodes to start a crawl with
MAX_CRAWL_NODES = 8
# Maximum number of find-requests a single crawl is allowed to make (excluding punctures)
MAX_CRAWL_REQUESTS = 24
# Maximum number of simultaneous outstanding find-requests per crawl
MAX_CRAWL_TASKS = 4

MAX_VALUES_IN_STORE = 8
MAX_VALUES_IN_FIND = 8
MAX_NODES_IN_FIND = 8

# Target number of nodes at which a key-value pair should be stored
TARGET_NODES = 8


async def gather_without_errors(*futures):
    results = await gather(*futures, return_exceptions=True)
    return [r for r in results if not isinstance(r, Exception)]


class Request(RandomNumberCache):
    """
    This request cache keeps track of all outstanding requests within the DHTCommunity.
    """
    def __init__(self, community, msg_type, node, params=None, consume_errors=False, timeout=5.0):
        super(Request, self).__init__(community.request_cache, msg_type)
        self.msg_type = msg_type
        self.node = node
        self.params = params
        self.future = Future()
        self.start_time = time.time()
        self.consume_errors = consume_errors
        self.timeout = timeout

    @property
    def timeout_delay(self):
        return self.timeout

    def on_timeout(self):
        if not self.future.done():
            self._logger.debug('Timeout for %s to %s', self.msg_type, self.node)
            self.node.failed += 1
            if not self.consume_errors:
                self.future.set_exception(DHTError('Timeout for {} to {}'.format(self.msg_type, self.node)))
            else:
                self.future.set_result(None)

    def on_complete(self):
        self.node.last_response = time.time()
        self.node.failed = 0
        self.node.rtt = time.time() - self.start_time


class Crawl:

    def __init__(self, target, routing_table, force_nodes=False, offset=0):
        self.target = target
        self.routing_table = routing_table

        nodes_closest = routing_table.closest_nodes(target, max_nodes=MAX_CRAWL_NODES)
        if not nodes_closest:
            raise DHTError('No nodes found in the routing table')

        # Keep a list of nodes that still need to be contacted: [(node_to_contact, node_to_puncture)]
        self.nodes_todo = [[n, None] for n in nodes_closest]
        self.nodes_tried = set()
        self.responses = []

        self.force_nodes = force_nodes
        self.offset = offset

    def add_response(self, sender, response):
        self.responses.append((sender, response))

        for index, node in enumerate(response.get('nodes', [])):
            if node in self.nodes_tried:
                continue

            # Only add nodes that are better than our current top-4
            if len(self.nodes_todo) >= 4 \
               and distance(node.id, self.target) > distance(self.nodes_todo[3][0].id, self.target):
                continue

            index_existing = next((i for i, t in enumerate(self.nodes_todo) if t[0] == node), None)
            if index_existing is not None:
                self.nodes_todo[index_existing][1] = None if index == 0 else sender
                continue

            self.nodes_todo.append([node, None if index == 0 else sender])
            self.nodes_todo.sort(key=lambda t: distance(t[0].id, self.target))

    @property
    def done(self):
        return len(self.nodes_tried) >= MAX_CRAWL_REQUESTS or not self.nodes_todo

    @property
    def cache_candidate(self):
        # Return closest node to the target that did not respond with values
        nodes_no_values = [sender for sender, response in self.responses if 'values' not in response]
        nodes_no_values.sort(key=lambda n: distance(n.id, self.target))
        return nodes_no_values[0] if nodes_no_values else None

    @property
    def values(self):
        # Merge all values received into one tuple. First pick the first value from each tuple, then the second, etc.
        value_responses = [response['values'] for _, response in self.responses if 'values' in response]
        values = sum(zip_longest(*value_responses), ())

        # Filter out duplicates while preserving order
        seen = set()
        return [v for v in values if v is not None and not (v in seen or seen.add(v))]

    @property
    def nodes(self):
        return sorted(self.nodes_tried, key=lambda n: distance(n.id, self.target))


class DHTCommunity(Community):
    """
    Community for storing/finding key-value pairs.
    """
    community_id = unhexlify('8d0be1845d74d175f178197cad001591d04d73cc')

    def __init__(self, my_peer, endpoint, network, *args, **kwargs):
        super().__init__(my_peer, endpoint, network, *args, **kwargs)
        self.network = Network()
        self.network.blacklist_mids.append(self.my_peer.mid)
        self.network.blacklist.extend(network.blacklist)

        self.storages = {}
        self.routing_tables = {}
        self.request_cache = RequestCache()
        self.tokens = {}
        self.token_secrets = deque(maxlen=2)

        # First call to token_maintenance should happen immediately, in case we get requests before it gets executed
        self.token_maintenance()

        self.register_task('token_maintenance', self.token_maintenance, interval=300)
        self.register_task('node_maintenance', self.node_maintenance, interval=60)

        # Register messages
        self.add_message_handler(PingRequestPayload, self.on_ping_request)
        self.add_message_handler(PingResponsePayload, self.on_ping_response)
        self.add_message_handler(StoreRequestPayload, self.on_store_request)
        self.add_message_handler(StoreResponsePayload, self.on_store_response)
        self.add_message_handler(FindRequestPayload, self.on_find_request)
        self.add_message_handler(FindResponsePayload, self.on_find_response)

        self.logger.info('DHT community initialized (peer mid %s)', hexlify(self.my_peer.mid))

    def get_serializer(self):
        serializer = super().get_serializer()
        serializer.add_packer('node-list', ListOf(NodePacker(serializer)))
        return serializer

    def get_available_strategies(self):
        return {'PingChurn': PingChurn}

    async def unload(self):
        await self.request_cache.shutdown()
        await super().unload()

    def get_address_class(self, node):
        return node.address.__class__ if node.address.__class__ != tuple else UDPv4Address

    def get_routing_table(self, node):
        address_cls = self.get_address_class(node)
        if address_cls not in self.routing_tables:
            self.routing_tables[address_cls] = RoutingTable(self.get_my_node_id(node))
        return self.routing_tables[address_cls]

    def get_storage(self, node):
        address_cls = self.get_address_class(node)
        if address_cls not in self.storages:
            self.storages[address_cls] = Storage()
        return self.storages[address_cls]

    def get_my_node_id(self, node):
        address_cls = self.get_address_class(node)
        address = self.my_peer.addresses.get(address_cls, self.my_estimated_wan)
        return calc_node_id(address, self.my_peer.mid)

    def get_requesting_node(self, peer):
        routing_table = self.get_routing_table(peer)
        node = Node(peer.key, peer.address)
        if routing_table.has(node.id) and routing_table.get(node.id).blocked:
            self.logger.debug('Too many queries\'s, dropping packet')
            return

        node = routing_table.add(node) or node
        node.last_queries.append(time.time())
        return node

    def introduction_request_callback(self, peer, dist, payload):
        self.on_node_discovered(peer.public_key.key_to_bin(), peer.address)

    def introduction_response_callback(self, peer, dist, payload):
        self.on_node_discovered(peer.public_key.key_to_bin(), peer.address)

    def on_node_discovered(self, public_key_bin, source_address):
        if isinstance(source_address, UDPv6Address) and UDPv6Address not in self.my_peer.addresses:
            return

        # Filter out trackers
        if source_address not in self.network.blacklist:
            node = Node(public_key_bin, source_address)
            routing_table = self.get_routing_table(Node(public_key_bin, source_address))
            existed = routing_table.has(node.id)
            rt_node = routing_table.add(node)

            if not existed and rt_node:
                self.logger.debug('Added node %s to the routing table', node)
                # Ping the node in order to determine RTT
                self.ping(rt_node)

    def ping(self, node):
        self.logger.debug('Pinging node %s', node)
        cache = Request(self, 'ping', node, consume_errors=True)
        self.request_cache.add(cache)
        self.ez_send(node, PingRequestPayload(cache.number))
        node.last_ping_sent = time.time()
        return cache.future

    @lazy_wrapper_wd(PingRequestPayload)
    def on_ping_request(self, peer, payload, data):
        self.logger.debug('Got ping-request from %s', peer.address)

        node = self.get_requesting_node(peer)
        if not node:
            return

        self.ez_send(peer, PingResponsePayload(payload.identifier))

    @lazy_wrapper_wd(PingResponsePayload)
    def on_ping_response(self, peer, payload, data):
        if not self.request_cache.has('ping', payload.identifier):
            self.logger.error('Got ping-response with unknown identifier, dropping packet')
            return

        self.logger.debug('Got ping-response from %s', peer.address)
        cache = self.request_cache.pop('ping', payload.identifier)
        cache.on_complete()
        if not cache.future.done():
            cache.future.set_result(cache.node)

    def serialize_value(self, data, sign=True):
        if sign:
            payload = SignedStrPayload(data, int(time.time()), self.my_peer.public_key.key_to_bin())
            return self._ez_pack(b'', DHT_ENTRY_STR_SIGNED, [payload], sig=True)
        payload = StrPayload(data)
        return self._ez_pack(b'', DHT_ENTRY_STR, [payload], sig=False)

    def unserialize_value(self, value):
        if value[0] == DHT_ENTRY_STR:
            payload, _ = self.serializer.unpack_serializable(StrPayload, value, offset=1)
            return payload.data, None, 0
        elif value[0] == DHT_ENTRY_STR_SIGNED:
            payload, _ = self.serializer.unpack_serializable(SignedStrPayload, value, offset=1)
            public_key = self.crypto.key_from_public_bin(payload.public_key)
            sig_len = self.crypto.get_signature_length(public_key)
            sig = value[-sig_len:]
            if self.crypto.is_valid_signature(public_key, value[:-sig_len], sig):
                return payload.data, payload.public_key, payload.version

    def add_value(self, key, value, storage, max_age=MAX_ENTRY_AGE):
        unserialized = self.unserialize_value(value)
        if unserialized:
            _, public_key, version = unserialized
            id_ = hashlib.sha1(public_key).digest() if public_key else None
            storage.put(key, value, id_=id_, version=version, max_age=max_age)
        else:
            self.logger.warning('Failed to store value %s', hexlify(value))

    async def store_value(self, key, data, sign=False):
        value = self.serialize_value(data, sign=sign)
        return await self._store(key, value)

    async def _store(self, key, value):
        if len(value) > MAX_ENTRY_SIZE:
            raise DHTError('Maximum length exceeded')

        nodes = await self.find_nodes(key)
        nodes = await self.store_on_nodes(key, [value], nodes[:TARGET_NODES])
        if len(nodes) < 1:
            raise DHTError('Failed to store value on DHT')
        return nodes

    @task
    async def store_on_nodes(self, key, values, nodes):
        if not nodes:
            raise DHTError('No nodes found for storing the key-value pairs')

        values = values[:MAX_VALUES_IN_STORE]

        # Check if we also need to store this key-value pair
        largest_distance = max([distance(node.id, key) for node in nodes])
        if len(nodes) < TARGET_NODES or distance(self.get_my_node_id(nodes[0]), key) < largest_distance:
            storage = self.get_storage(nodes[0])
            for value in reversed(values):
                self.add_value(key, value, storage)

        now = time.time()
        futures = []
        for node in nodes:
            if node.id in self.tokens and self.tokens[node.id][0] + TOKEN_EXPIRATION_TIME > now:
                cache = Request(self, 'store', node)
                self.request_cache.add(cache)
                futures.append(cache.future)
                self.ez_send(node, StoreRequestPayload(cache.number, self.tokens[node.id][1], key, values))
            else:
                self.logger.debug('Not sending store-request to %s (no token available)', node)

        if not futures:
            raise DHTError('Value was not stored')
        return await gather_without_errors(*futures)

    @lazy_wrapper(StoreRequestPayload)
    def on_store_request(self, peer, payload):
        self.logger.debug('Got store-request from %s', peer.address)

        node = self.get_requesting_node(peer)
        if not node:
            return
        if any([len(value) > MAX_ENTRY_SIZE for value in payload.values]):
            self.logger.warning('Maximum length of value exceeded, dropping packet.')
            return
        if len(payload.values) > MAX_VALUES_IN_STORE:
            self.logger.warning('Too many values, dropping packet.')
            return
        # Note that even though we are preventing spoofing of source_address (by checking the token),
        # the value that is to be stored isn't checked. This should be done at a higher level.
        if not self.check_token(node, payload.token):
            self.logger.warning('Bad token, dropping packet.')
            return

        # How many nodes (that we know of) are closer to this value?
        num_closer = 0
        for node in self.get_routing_table(node).closest_nodes(payload.target, max_nodes=20):
            if distance(node.id, payload.target) < distance(self.get_my_node_id(node), payload.target):
                num_closer += 1

        # To prevent over-caching, the expiration time of an entry depends on the number
        # of nodes that are closer than us.
        max_age = MAX_ENTRY_AGE // 2 ** max(0, num_closer - TARGET_NODES + 1)
        for value in payload.values:
            self.add_value(payload.target, value, self.get_storage(node), max_age)

        self.ez_send(peer, StoreResponsePayload(payload.identifier))

    @lazy_wrapper(StoreResponsePayload)
    def on_store_response(self, peer, payload):
        if not self.request_cache.has('store', payload.identifier):
            self.logger.error('Got store-response with unknown identifier, dropping packet')
            return

        self.logger.debug('Got store-response from %s', peer.address)
        cache = self.request_cache.pop('store', payload.identifier)
        cache.on_complete()
        if not cache.future.done():
            cache.future.set_result(cache.node)

    def _send_find_request(self, node, target, force_nodes, offset=0):
        cache = Request(self, 'find', node, [force_nodes], consume_errors=True, timeout=2.0)
        self.request_cache.add(cache)
        self.ez_send(node, FindRequestPayload(cache.number, self.my_estimated_lan, target, offset, force_nodes))
        return cache.future

    async def _contact_node(self, crawl, node, puncture_node):
        if puncture_node:
            await self._send_find_request(puncture_node, node.id, crawl.force_nodes)
        result = await self._send_find_request(node, crawl.target, crawl.force_nodes, crawl.offset)
        if result:
            crawl.routing_table.add(node)
            crawl.add_response(node, result)

    async def _find(self, crawl, debug=False):
        tasks = set()
        while True:
            # Keep running tasks until work is done.
            while not crawl.done and len(tasks) < MAX_CRAWL_TASKS:
                node, puncture_node = crawl.nodes_todo.pop(0)
                tasks.add(self.register_anonymous_task('contact_node', self._contact_node, crawl, node, puncture_node))
                # Add to nodes_tried immediately to prevent sending multiple find-requests to the same node.
                crawl.nodes_tried.add(node)
            if not tasks:
                break
            _, tasks = await wait(tasks, return_when=FIRST_COMPLETED)

        if crawl.force_nodes:
            return crawl.nodes

        cache_candidate = crawl.cache_candidate
        values = crawl.values

        if cache_candidate and values:
            # Store the key-value pair on the most recently visited node that
            # did not have it (for caching purposes).
            self.store_on_nodes(crawl.target, values, [cache_candidate])

        if debug:
            return self.post_process_values(values), crawl
        return self.post_process_values(values)

    def post_process_values(self, values):
        # Unpack values and filter out duplicates
        unpacked = defaultdict(list)
        for value in values:
            unserialized = self.unserialize_value(value)
            if unserialized:
                data, public_key, version = unserialized
                unpacked[public_key].append((version, data))

        results = []

        # Signed data
        for public_key, data_list in unpacked.items():
            if public_key is not None:
                results.append((max(data_list, key=lambda t: t[0])[1], public_key))

        # Unsigned data
        for data in unpacked[None]:
            results.append((data[1], None))

        return results

    async def find(self, target, force_nodes, offset, debug):
        futures = []
        for routing_table in self.routing_tables.values():
            crawl = Crawl(target, routing_table, force_nodes=force_nodes, offset=offset)
            futures.append(self._find(crawl, debug=debug))
        results = await gather(*futures)
        if debug:
            return sum([r[0] for r in results], []), [r[1] for r in results]
        return sum(results, [])

    async def find_values(self, target, offset=0, debug=False):
        return await self.find(target, False, offset, debug)

    async def find_nodes(self, target, debug=False):
        return await self.find(target, True, 0, debug)

    @lazy_wrapper(FindRequestPayload)
    def on_find_request(self, peer, payload):
        self.logger.debug('Got find-request for %s from %s', hexlify(payload.target), peer.address)

        node = self.get_requesting_node(peer)
        if not node:
            return

        nodes = []
        storage = self.get_storage(node)
        values = storage.get(payload.target, starting_point=payload.offset, limit=MAX_VALUES_IN_FIND) \
            if not payload.force_nodes else []

        if payload.force_nodes or not values:
            routing_table = self.get_routing_table(node)
            nodes = routing_table.closest_nodes(payload.target, exclude_node=node, max_nodes=MAX_NODES_IN_FIND)
            # Send puncture request to the closest node
            if nodes:
                packet = self.create_puncture_request(payload.lan_address, peer.address, payload.identifier)
                self.endpoint.send(nodes[0].address, packet)

        self.ez_send(peer, FindResponsePayload(payload.identifier, self.generate_token(node), values, nodes))

    @lazy_wrapper(FindResponsePayload)
    def on_find_response(self, peer, payload):
        if not self.request_cache.has('find', payload.identifier):
            self.logger.error('Got find-response with unknown identifier, dropping packet')
            return

        self.logger.debug('Got find-response from %s', peer.address)
        cache = self.request_cache.pop('find', payload.identifier)
        cache.on_complete()

        self.tokens[cache.node.id] = (time.time(), payload.token)

        if cache.future.done():
            # The errback must already have been called (due to a timeout)
            return
        elif cache.params[0]:
            cache.future.set_result({'nodes': payload.nodes})
        else:
            cache.future.set_result({'values': payload.values} if payload.values else {'nodes': payload.nodes})

    async def node_maintenance(self):
        # Determine which buckets to refresh
        refresh_todo = {}
        now = time.time()
        for routing_table in self.routing_tables.values():
            for bucket in routing_table.trie.values():
                if now - bucket.last_changed > 15 * 60:
                    refresh_todo[bucket.prefix_id] = refresh_todo.get(bucket.prefix_id, [])
                    refresh_todo[bucket.prefix_id].append(bucket)

        # For every prefix perform a refresh.
        # We only have to perform a single refresh for each prefix because find_values will crawl both IPv4 and IPv6.
        for buckets in refresh_todo.values():
            try:
                await self.find_values(buckets[0].generate_id())
            except DHTError:
                pass

            for bucket in buckets:
                bucket.last_changed = now

    def token_maintenance(self):
        self.token_secrets.append(os.urandom(16))

        # Cleanup old tokens
        now = time.time()
        for node_id, (ts, _) in list(self.tokens.items()):
            if now > ts + TOKEN_EXPIRATION_TIME:
                self.tokens.pop(node_id, None)

    def generate_token(self, node):
        return hashlib.sha1(str(node).encode() + self.token_secrets[-1]).digest()

    def check_token(self, node, token):
        return any([hashlib.sha1(str(node).encode() + secret).digest() == token for secret in self.token_secrets])



import random
from asyncio import get_event_loop

from ...messaging.interfaces.endpoint import Endpoint, EndpointListener
from ...messaging.interfaces.udp.endpoint import UDPv4Address, UDPv6Address

internet = {}


class MockEndpoint(Endpoint):

    def __init__(self, lan_address, wan_address):
        super(MockEndpoint, self).__init__()
        internet[lan_address] = self
        internet[wan_address] = self

        self.lan_address = lan_address
        self.wan_address = wan_address

        self._port = self.lan_address[1]
        self._open = False

    def assert_open(self):
        assert self._open

    def is_open(self):
        return self._open

    def get_address(self):
        return self.wan_address

    def send(self, socket_address, packet):
        if not self.is_open():
            return
        if socket_address in internet:
            # For the unit tests we handle messages in separate asyncio tasks to prevent infinite recursion.
            ep = internet[socket_address]
            get_event_loop().call_soon(ep.notify_listeners, (self.wan_address, packet))
        else:
            raise AssertionError("Received data from unregistered address %s" % repr(socket_address))

    def open(self):
        self._open = True

    def close(self, timeout=0.0):
        self._open = False

    def reset_byte_counters(self):
        pass


class AddressTester(EndpointListener):

    def on_packet(self, packet):
        pass

    def is_lan(self, address):
        if isinstance(address, UDPv4Address):
            return self._address_is_lan_without_netifaces(address)
        return False


class AutoMockEndpoint(MockEndpoint):

    ADDRESS_TYPE = "UDPv4Address"

    def __init__(self):
        self._open = False
        super(AutoMockEndpoint, self).__init__(self._generate_unique_address(), self._generate_unique_address())
        self._port = 0

    def _generate_address(self):
        if self.ADDRESS_TYPE == "UDPv4Address":
            b0 = random.randint(0, 255)
            b1 = random.randint(0, 255)
            b2 = random.randint(0, 255)
            b3 = random.randint(0, 255)
            port = random.randint(0, 65535)

            return UDPv4Address('%d.%d.%d.%d' % (b0, b1, b2, b3), port)
        elif self.ADDRESS_TYPE == "UDPv6Address":
            b0 = random.randint(0, 65535)
            b1 = random.randint(0, 65535)
            b2 = random.randint(0, 65535)
            b3 = random.randint(0, 65535)
            b4 = random.randint(0, 65535)
            b5 = random.randint(0, 65535)
            b6 = random.randint(0, 65535)
            b7 = random.randint(0, 65535)
            port = random.randint(0, 65535)

            return UDPv6Address('%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x' % (b0, b1, b2, b3, b4, b5, b6, b7), port)
        else:
            raise RuntimeError("Illegal address type specified: " + repr(self.ADDRESS_TYPE))

    def _is_lan(self, address):
        """
        Avoid false positives for the actual machine's lan.
        """
        self._port = address[1]
        address_tester = AddressTester(self)
        return address_tester.is_lan(address[0])

    def _generate_unique_address(self):
        address = self._generate_address()

        while address in internet or self._is_lan(address):
            address = self._generate_address()

        return address


class MockEndpointListener(EndpointListener):

    def __init__(self, endpoint, main_thread=False):
        super(MockEndpointListener, self).__init__(endpoint, main_thread)

        self.received_packets = []

        endpoint.add_listener(self)

    def on_packet(self, packet):
        self.received_packets.append(packet)



import abc
import logging
import socket
import struct
import threading

try:
    # Especially on Android netifaces may fail.
    # Generally, we also allow users not to have netifaces installed.
    import netifaces
except ImportError:
    netifaces = None


class Endpoint(metaclass=abc.ABCMeta):
    """
    Interface for sending messages over the Internet.
    """

    def __init__(self, prefixlen=22):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._listeners = []
        self._prefix_map = {}
        self.prefixlen = prefixlen
        self.listener_update_lock = threading.RLock()

    def add_listener(self, listener):
        """
        Add an EndpointListener to our listeners.

        :raises: IllegalEndpointListenerError if the provided listener is not an EndpointListener
        """
        if not isinstance(listener, EndpointListener):
            raise IllegalEndpointListenerError(listener)
        with self.listener_update_lock:
            self._listeners.append(listener)
            for prefix in self._prefix_map:
                self._prefix_map[prefix].append(listener)

    def add_prefix_listener(self, listener, prefix):
        """
        Add an EndpointListener to our listeners, only triggers on packets with a specific prefix.

        :raises: IllegalEndpointListenerError if the provided listener is not an EndpointListener
        """
        if not isinstance(listener, EndpointListener):
            raise IllegalEndpointListenerError(listener)
        if not len(prefix) == self.prefixlen:
            raise RuntimeError("Tried to register a prefix of length %d, required to be of length %d!"
                               % (len(prefix), self.prefixlen))
        with self.listener_update_lock:
            self._prefix_map[prefix] = self._prefix_map.get(prefix, []) + [listener] + self._listeners

    def remove_listener(self, listener):
        """
        Remove a listener from our listeners, if it is registered.
        """
        with self.listener_update_lock:
            self._listeners = [l for l in self._listeners if l != listener]
            new_prefix_map = {}
            for prefix in self._prefix_map:
                listeners = [l for l in self._prefix_map[prefix] if l != listener]
                if set(listeners) != set(self._listeners):
                    new_prefix_map[prefix] = listeners
            self._prefix_map = new_prefix_map

    def _deliver_later(self, listener, packet):
        """
        Ensure that the listener is still loaded when delivering the packet later.
        """

        if self.is_open() and (packet[1][:self.prefixlen] in self._prefix_map or listener in self._listeners):
            listener.on_packet(packet)

    def notify_listeners(self, packet):
        """
        Send data to all listeners.

        :param data: the data to send to all listeners.
        """
        prefix = packet[1][:self.prefixlen]
        listeners = self._prefix_map.get(prefix, self._listeners)
        for listener in listeners:
            # TODO: Respect listener.use_main_thread:
            self._deliver_later(listener, packet)

    @abc.abstractmethod
    def assert_open(self):
        pass

    @abc.abstractmethod
    def is_open(self):
        pass

    @abc.abstractmethod
    def get_address(self):
        pass

    @abc.abstractmethod
    def send(self, socket_address, packet):
        pass

    @abc.abstractmethod
    def open(self):
        pass

    @abc.abstractmethod
    def close(self):
        pass

    @abc.abstractmethod
    def reset_byte_counters(self):
        pass


class EndpointListener(metaclass=abc.ABCMeta):
    """
    Handler for messages coming in through an Endpoint.
    """

    def __init__(self, endpoint, main_thread=True):
        """
        Create a new listener.

        :param main_thread: run the callback of this listener on the main thread.
        """
        self._use_main_thread = main_thread

        self.endpoint = endpoint

        self._netifaces_failed = netifaces is None
        self._local_interfaces = []
        self._my_estimated_lan = None
        self.my_estimated_wan = self.my_estimated_lan

    @property
    def my_estimated_lan(self):
        """
        Estimate our LAN IPv4 address and port.

        If the endpoint is closed this returns ("0.0.0.0", 0).
        If the endpoint is open and we have no idea what our address is, attempt to estimate it.
        Otherwise, return the current value of the estimated LAN address and port.
        """
        if not self.endpoint.is_open():
            return "0.0.0.0", 0
        if self._my_estimated_lan is None:
            self._my_estimated_lan = (self._get_lan_address(True)[0], self.endpoint.get_address()[1])
        return self._my_estimated_lan

    @my_estimated_lan.setter
    def my_estimated_lan(self, value):
        self._my_estimated_lan = value

    @property
    def use_main_thread(self):
        """
        Does the callback of this listener need to be executed on the main thread.
        """
        return self._use_main_thread

    @abc.abstractmethod
    def on_packet(self, packet):
        """
        Callback for when data is received on this endpoint.

        :param packet: the received packet, in (source, binary string) format.
        """
        pass

    @staticmethod
    def _get_interface_addresses():
        """
        Yields Interface instances for each available AF_INET interface found.

        An Interface instance has the following properties:
        - name          (i.e. "eth0")
        - address       (i.e. "10.148.3.254")
        - netmask       (i.e. "255.255.255.0")
        - broadcast     (i.e. "10.148.3.255")
        """

        class Interface(object):

            def __init__(self, name, address, netmask, broadcast):
                self.name = name
                self.address = address
                self.netmask = netmask
                self.broadcast = broadcast
                self._l_address, = struct.unpack_from(">L", socket.inet_aton(address))
                self._l_netmask, = struct.unpack_from(">L", socket.inet_aton(netmask))

            def __contains__(self, address):
                assert isinstance(address, str), type(address)
                l_address, = struct.unpack_from(">L", socket.inet_aton(address))
                return (l_address & self._l_netmask) == (self._l_address & self._l_netmask)

            def __str__(self):
                return "<{self.__class__.__name__} \"{self.name}\" addr:{self.address} mask:{self.netmask}>".format(
                    self=self)

            def __repr__(self):
                return "<{self.__class__.__name__} \"{self.name}\" addr:{self.address} mask:{self.netmask}>".format(
                    self=self)

        try:
            for interface in netifaces.interfaces():
                try:
                    addresses = netifaces.ifaddresses(interface)

                except ValueError:
                    # some interfaces are given that are invalid, we encountered one called ppp0
                    pass

                else:
                    for option in addresses.get(netifaces.AF_INET, []):
                        try:
                            # On Windows netifaces currently returns IP addresses as unicode,
                            # and on *nix it returns str. So, we convert any unicode objects to str.
                            # Python 3 port update: In Python 3 everything is unicode and we should do nothing.
                            # On Python 2 the above can happen: instead of checking for unicode, we check for `not str`.
                            unicode_to_str = lambda s: s.encode('utf-8') if s and not isinstance(s, str) else s
                            yield Interface(interface,
                                            unicode_to_str(option.get("addr")),
                                            unicode_to_str(option.get("netmask")),
                                            unicode_to_str(option.get("broadcast")))

                        except TypeError:
                            # some interfaces have no netmask configured, causing a TypeError when
                            # trying to unpack _l_netmask
                            pass
        except OSError as e:
            logger = logging.getLogger("dispersy")
            logger.warning("failed to check network interfaces, error was: %r", e)

    def _address_in_subnet(self, address, subnet):
        """
        Checks whether a given address is in a given subnet
        :param address: an ip v4 address as a string formatted as four pairs of decimals separated by dots
        :param subnet: a tuple consisting of the main address of the subnet formatted as above, and the subnet formatted as
        an int with the number of significant bits in the address.
        :return: True if the address is in the subnet, False otherwise
        """
        address = struct.unpack_from(">L", socket.inet_aton(address))[0]
        (subnet_main, netmask) = subnet
        subnet_main = struct.unpack_from(">L", socket.inet_aton(subnet_main))[0]
        address >>= 32 - netmask
        subnet_main >>= 32 - netmask
        return address == subnet_main

    def _address_is_lan_without_netifaces(self, address):
        """
        Checks if the given ip address is either our own address or in one of the subnet defined for local network usage
        :param address: ip v4 address to be checked
        :return: True if the address is a lan address, False otherwise
        """
        if address == self.get_lan_address_without_netifaces():
            return True
        return self.address_in_lan_subnets(address)

    def address_in_lan_subnets(self, address):
        lan_subnets = (("192.168.0.0", 16),
                       ("172.16.0.0", 12),
                       ("10.0.0.0", 8))
        return any(self._address_in_subnet(address, subnet) for subnet in lan_subnets)

    def address_is_lan(self, address):
        if self._netifaces_failed:
            return self._address_is_lan_without_netifaces(address)
        else:
            return any(address in interface for interface in self._local_interfaces)

    def get_lan_address_without_netifaces(self):
        """
        # Get the local ip address by creating a socket for a (random) internet ip
        :return: the local ip address
        """
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect(("192.0.2.0", 80))  # TEST-NET-1, guaranteed to not be connected => no callbacks
            local_ip = s.getsockname()[0]
            s.close()
            return local_ip
        except socket.error:
            return "0.0.0.0"

    def get_ipv6_address(self):
        for interface in netifaces.interfaces():
            try:
                addresses = netifaces.ifaddresses(interface)
                for option in addresses.get(netifaces.AF_INET6, []):
                    try:
                        if option.get("addr") == "::1":
                            continue
                        unicode_to_str = lambda s: s.encode('utf-8') if s and not isinstance(s, str) else s
                        return unicode_to_str(option.get("addr").split('%')[0])  # Platform-specific may have "%eth0"
                    except TypeError:
                        pass
            except OSError as e:
                logger = logging.getLogger("dispersy")
                logger.warning("failed to check network interfaces, error was: %r", e)

    def _get_lan_address(self, bootstrap=False):
        """
        Attempt to get the newest lan ip of this machine, preferably with netifaces, but use the fallback if it fails
        :return: lan address
        """
        if self._netifaces_failed:
            return self.get_lan_address_without_netifaces(), self.endpoint.get_address()[1]
        else:
            self._local_interfaces = list(self._get_interface_addresses())
            interface = self._guess_lan_address(self._local_interfaces)
            return (interface.address if interface else self.get_lan_address_without_netifaces()), \
                   (0 if bootstrap else self.endpoint.get_address()[1])

    def _guess_lan_address(self, interfaces, default=None):
        """
        Chooses the most likely Interface instance out of INTERFACES to use as our LAN address.

        INTERFACES can be obtained from _get_interface_addresses()
        DEFAULT is used when no appropriate Interface can be found
        """
        assert isinstance(interfaces, list), type(interfaces)
        blacklist = ["127.0.0.1", "0.0.0.0", "255.255.255.255"]

        # prefer interfaces where we have a broadcast address
        for interface in interfaces:
            if interface.broadcast and interface.address and interface.address not in blacklist:
                return interface

        # Exception for virtual machines/containers
        for interface in interfaces:
            if interface.address and interface.address not in blacklist:
                return interface

        self._netifaces_failed = True
        return default


class IllegalEndpointListenerError(RuntimeError):
    """
    Exception raised when an EndpointListener instance was expected, but not supplied.
    """

    def __init__(self, other):
        message = '%s is not an instance of %s' % (type(other), str(EndpointListener.__name__))
        super(IllegalEndpointListenerError, self).__init__(message)


class EndpointClosedException(Exception):
    """
    Exception raised when an endpoint is expected to be open, but is closed.
    """

    def __init__(self, endpoint):
        super(EndpointClosedException, self).__init__('%s is unexpectedly closed' % type(endpoint))


class DataTooBigException(Exception):
    """
    Exception raised when the data being sent exceeds the maximum size.
    """

    def __init__(self, size, max_size):
        super(DataTooBigException, self).__init__('Tried to send packet of size %s > MAX_SIZE(%d)' % (size, max_size))


class IllegalDestination(Exception):
    """
    Exception raised when trying to send to the 0 address.
    """

    def __init__(self):
        super(IllegalDestination, self).__init__('Attempted to send a message to 0.0.0.0:0.')



from ..base import TestBase
from ...messaging.lazy_payload import VariablePayload, vp_compile
from ...messaging.payload import Payload
from ...messaging.serialization import default_serializer


class A(VariablePayload):
    """
    A basic VariablePayload.
    """
    format_list = ['I', 'H']
    names = ["a", "b"]


@vp_compile
class CompiledA(A):
    pass


class BitsPayload(VariablePayload):
    """
    An unbalanced VariablePayload.
    """
    format_list = ['bits']
    names = ['flag0', 'flag1', 'flag2', 'flag3', 'flag4', 'flag5', 'flag6', 'flag7']


@vp_compile
class CompiledBitsPayload(BitsPayload):
    pass


class B(VariablePayload):
    """
    A VariablePayload with a nested Payload.
    """
    format_list = [A]
    names = ["a"]


@vp_compile
class CompiledB(B):
    pass


@vp_compile
class CompiledBAlt(B):
    format_list = [CompiledA]


class C(A):
    """
    A VariablePayload with inherited fields.
    """
    format_list = A.format_list + ['B']
    names = A.names + ['c']


@vp_compile
class CompiledC(C):
    pass


@vp_compile
class CompiledCAlt(CompiledA):
    format_list = A.format_list + ['B']
    names = A.names + ['c']


class OldA(Payload):
    format_list = ["I", "H"]

    def __init__(self, a, b):
        self.a = a
        self.b = b

    def to_pack_list(self):
        return [("I", self.a),
                ("H", self.b)]

    @classmethod
    def from_unpack_list(cls, *args):
        return OldA(*args)  # pylint: disable=E1120


class D(VariablePayload):
    format_list = ["I"]
    names = ["a"]

    def fix_pack_a(self, value):
        return value + 1

    @classmethod
    def fix_unpack_a(cls, value):
        return value - 1


@vp_compile
class CompiledD(D):
    pass


class NewC(VariablePayload, OldA):
    """
    A VariablePayload with inherited fields.
    """
    format_list = OldA.format_list + ['B']
    names = ['a', 'b', 'c']


@vp_compile
class CompiledNewC(NewC):
    pass


class E(VariablePayload):
    """
    A VariablePayload with a list of payloads.
    """
    format_list = [[A]]
    names = ["list_of_A"]


class TestVariablePayload(TestBase):

    def _pack_and_unpack(self, payload, instance):
        """
        Serialize and unserialize an instance of payload.
        :param payload: the payload class to serialize for
        :type payload: type(Payload)
        :param instance: the payload instance to serialize
        :type instance: Payload
        :return: the repacked instance
        """
        serialized = default_serializer.pack_serializable(instance)
        deserialized, _ = default_serializer.unpack_serializable(payload, serialized)  # pylint: disable=E0632
        return deserialized

    def test_base_unnamed(self):
        """
        Check if the wrapper returns the payload correctly with unnamed arguments.
        """
        a = A(42, 1337)

        deserialized = self._pack_and_unpack(A, a)

        self.assertEqual(a.a, 42)
        self.assertEqual(a.b, 1337)
        self.assertEqual(deserialized.a, 42)
        self.assertEqual(deserialized.b, 1337)

    def test_base_unnamed_compiled(self):
        """
        Check if the wrapper returns the payload correctly with unnamed arguments, compiled.
        """
        a = CompiledA(42, 1337)

        deserialized = self._pack_and_unpack(CompiledA, a)

        self.assertEqual(a.a, 42)
        self.assertEqual(a.b, 1337)
        self.assertEqual(deserialized.a, 42)
        self.assertEqual(deserialized.b, 1337)

    def test_base_named(self):
        """
        Check if the wrapper returns the payload correctly with named arguments.
        """
        a = A(b=1337, a=42)

        deserialized = self._pack_and_unpack(A, a)

        self.assertEqual(a.a, 42)
        self.assertEqual(a.b, 1337)
        self.assertEqual(deserialized.a, 42)
        self.assertEqual(deserialized.b, 1337)

    def test_base_named_compiled(self):
        """
        Check if the wrapper returns the payload correctly with named arguments, compiled.
        """
        a = CompiledA(b=1337, a=42)

        deserialized = self._pack_and_unpack(CompiledA, a)

        self.assertEqual(a.a, 42)
        self.assertEqual(a.b, 1337)
        self.assertEqual(deserialized.a, 42)
        self.assertEqual(deserialized.b, 1337)

    def test_bits_payload(self):
        """
        Check if unpacked BitPayload works correctly.
        """
        bpl = BitsPayload(False, True, False, True, False, True, False, True)

        deserialized = self._pack_and_unpack(BitsPayload, bpl)

        self.assertEqual(deserialized.flag0, False)
        self.assertEqual(deserialized.flag1, True)
        self.assertEqual(deserialized.flag2, False)
        self.assertEqual(deserialized.flag3, True)
        self.assertEqual(deserialized.flag4, False)
        self.assertEqual(deserialized.flag5, True)
        self.assertEqual(deserialized.flag6, False)
        self.assertEqual(deserialized.flag7, True)

    def test_bits_payload_compiled(self):
        """
        Check if unpacked compiled BitPayload works correctly.
        """
        bpl = CompiledBitsPayload(False, True, False, True, False, True, False, True)

        deserialized = self._pack_and_unpack(BitsPayload, bpl)

        self.assertEqual(deserialized.flag0, False)
        self.assertEqual(deserialized.flag1, True)
        self.assertEqual(deserialized.flag2, False)
        self.assertEqual(deserialized.flag3, True)
        self.assertEqual(deserialized.flag4, False)
        self.assertEqual(deserialized.flag5, True)
        self.assertEqual(deserialized.flag6, False)
        self.assertEqual(deserialized.flag7, True)

    def test_inheritance(self):
        """
        Check if the wrapper allows for nested payloads.
        """
        a = A(1, 2)
        b = B(a)

        deserialized = self._pack_and_unpack(B, b)

        self.assertEqual(b.a.a, deserialized.a.a)
        self.assertEqual(b.a.b, deserialized.a.b)
        self.assertIsInstance(deserialized, B)
        self.assertIsInstance(deserialized.a, A)

    def test_inheritance_uncompiled_compiled(self):
        """
        Check if the wrapper allows for nested payloads, compiled.
        """
        a = CompiledA(1, 2)
        b = CompiledB(a)

        deserialized = self._pack_and_unpack(CompiledB, b)

        self.assertEqual(b.a.a, deserialized.a.a)
        self.assertEqual(b.a.b, deserialized.a.b)
        self.assertIsInstance(deserialized, CompiledB)
        self.assertIsInstance(deserialized.a, A)

    def test_inheritance_compiled_compiled(self):
        """
        Check if the wrapper allows for compiled nested payloads, compiled.
        """
        a = CompiledA(1, 2)
        b = CompiledBAlt(a)

        deserialized = self._pack_and_unpack(CompiledBAlt, b)

        self.assertEqual(b.a.a, deserialized.a.a)
        self.assertEqual(b.a.b, deserialized.a.b)
        self.assertIsInstance(deserialized, CompiledBAlt)
        self.assertIsInstance(deserialized.a, CompiledA)

    def test_subclass(self):
        """
        Check if the wrapper allows for subclasses.
        """
        c = C(1, c=3, b=2)

        deserialized = self._pack_and_unpack(C, c)

        self.assertEqual(c.a, deserialized.a)
        self.assertEqual(c.b, deserialized.b)
        self.assertEqual(c.c, deserialized.c)
        self.assertIsInstance(deserialized, C)
        self.assertIsInstance(deserialized, A)

    def test_subclass_uncompiled_compiled(self):
        """
        Check if the wrapper allows for subclasses, compiled.
        """
        c = CompiledC(1, c=3, b=2)

        deserialized = self._pack_and_unpack(CompiledC, c)

        self.assertEqual(c.a, deserialized.a)
        self.assertEqual(c.b, deserialized.b)
        self.assertEqual(c.c, deserialized.c)
        self.assertIsInstance(deserialized, CompiledC)
        self.assertIsInstance(deserialized, A)

    def test_subclass_compiled_compiled(self):
        """
        Check if the wrapper allows for compiled subclasses, compiled.
        """
        c = CompiledCAlt(1, c=3, b=2)

        deserialized = self._pack_and_unpack(CompiledCAlt, c)

        self.assertEqual(c.a, deserialized.a)
        self.assertEqual(c.b, deserialized.b)
        self.assertEqual(c.c, deserialized.c)
        self.assertIsInstance(deserialized, CompiledCAlt)
        self.assertIsInstance(deserialized, CompiledA)

    def test_old_subclass(self):
        """
        Check if the wrapper allows for subclasses from old-style Payloads.
        """
        c = NewC(1, c=3, b=2)

        deserialized = self._pack_and_unpack(NewC, c)

        self.assertEqual(c.a, deserialized.a)
        self.assertEqual(c.b, deserialized.b)
        self.assertEqual(c.c, deserialized.c)
        self.assertIsInstance(deserialized, NewC)
        self.assertIsInstance(deserialized, OldA)

    def test_old_subclass_compiled(self):
        """
        Check if the wrapper allows for subclasses from old-style Payloads, compiled.
        """
        c = CompiledNewC(1, c=3, b=2)

        deserialized = self._pack_and_unpack(CompiledNewC, c)

        self.assertEqual(c.a, deserialized.a)
        self.assertEqual(c.b, deserialized.b)
        self.assertEqual(c.c, deserialized.c)
        self.assertIsInstance(deserialized, CompiledNewC)
        self.assertIsInstance(deserialized, OldA)

    def test_custom_pack(self):
        """
        Check if the wire-format manipulation rules are applied correctly.
        """
        d = D(0)

        serialized = default_serializer.pack_serializable(d)
        deserialized = self._pack_and_unpack(D, d)

        self.assertEqual(d.a, 0)
        self.assertEqual(deserialized.a, 0)
        self.assertEqual(serialized, b'\x00\x00\x00\x01')

    def test_custom_pack_compiled(self):
        """
        Check if the wire-format manipulation rules are applied correctly, compiled.
        """
        d = CompiledD(0)

        serialized = default_serializer.pack_serializable(d)
        deserialized = self._pack_and_unpack(CompiledD, d)

        self.assertEqual(d.a, 0)
        self.assertEqual(deserialized.a, 0)
        self.assertEqual(serialized, b'\x00\x00\x00\x01')

    def test_payload_list(self):
        """
        Check if unpacked payload lists works correctly.
        """
        e = E([A(1, 2), A(3, 4)])

        deserialized = self._pack_and_unpack(E, e)

        self.assertEqual(e.list_of_A[0].a, deserialized.list_of_A[0].a)
        self.assertEqual(e.list_of_A[0].b, deserialized.list_of_A[0].b)
        self.assertEqual(e.list_of_A[1].a, deserialized.list_of_A[1].a)
        self.assertEqual(e.list_of_A[1].b, deserialized.list_of_A[1].b)



from base64 import decodebytes, encodebytes
from binascii import hexlify
from math import ceil

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric.utils import encode_dss_signature

from .. import NEW_CRYPTOGRAPHY_SIGN_VERSION
from ...keyvault.keys import PublicKey


class M2CryptoPK(PublicKey):

    def __init__(self, ec_pub=None, keystring=None):
        """
        Create a new M2Crypto public key. Optionally load it from a string representation or
        using a public key.

        :param ec_pub: load the pk from a PubicKey object
        :param keystring: load the pk from this string (see key_to_bin())
        """
        if ec_pub:
            self.ec = ec_pub
        elif keystring:
            self.ec = self.key_from_pem(b"-----BEGIN PUBLIC KEY-----\n%s-----END PUBLIC KEY-----\n" %
                                        encodebytes(keystring))

    def pem_to_bin(self, pem):
        """
        Convert a key in the PEM format into a key in the binary format.
        @note: Encrypted pem's are NOT supported and will silently fail.
        """
        return decodebytes(b"".join(pem.split(b"\n")[1:-2]))

    def key_to_pem(self):
        "Convert a key to the PEM format."
        return self.ec.public_bytes(encoding=serialization.Encoding.PEM,
                                    format=serialization.PublicFormat.SubjectPublicKeyInfo)

    def key_from_pem(self, pem):
        "Get the EC from a public PEM."
        return serialization.load_pem_public_key(pem, backend=default_backend())

    def key_to_bin(self):
        """
        Get the string representation of this key.
        """
        return self.pem_to_bin(self.key_to_pem())

    def get_signature_length(self):
        """
        Returns the length, in bytes, of each signature made using EC.
        """
        return int(ceil(self.ec.curve.key_size / 8.0)) * 2

    def verify(self, signature, msg):
        """
        Verify whether a given signature is correct for a message.

        :param signature: the given signature
        :param msg: the given message
        """
        length = len(signature) // 2
        r = signature[:length].lstrip(b"\x00")  # remove all "\x00" prefixes
        # prepend "\x00" when the most significant bit is set
        if r[0] & 128:
            r = b"\x00" + r

        s = signature[length:].lstrip(b"\x00")  # remove all "\x00" prefixes
        # prepend "\x00" when the most significant bit is set
        if s[0] & 128:
            s = b"\x00" + s
        # turn back into int
        r = int(hexlify(r), 16)
        s = int(hexlify(s), 16)
        # verify
        try:
            if NEW_CRYPTOGRAPHY_SIGN_VERSION:
                self.pub().ec.verify(encode_dss_signature(r, s), msg, ec.ECDSA(hashes.SHA1()))
            else:
                self.pub().ec.verifier(encode_dss_signature(r, s), ec.ECDSA(hashes.SHA1()))
            return True
        except InvalidSignature:
            return False



from ..base import TestBase
from ...keyvault.crypto import default_eccrypto


class TestSignatures(TestBase):
    """
    Test whether signatures can be created and then decoded correctly.
    """

    def setUp(self):
        super(TestSignatures, self).setUp()
        self.ec = default_eccrypto
        self.data = bytes(range(256))

    def test_vlow(self):
        """
        Check if very-low security keys generate a valid signature.
        """
        key = self.ec.generate_key(u"very-low")

        signature = key.signature(self.data)

        self.assertTrue(self.ec.is_valid_signature(key.pub(), self.data, signature))

    def test_low(self):
        """
        Check if low security keys generate a valid signature.
        """
        key = self.ec.generate_key(u"low")

        signature = key.signature(self.data)

        self.assertTrue(self.ec.is_valid_signature(key.pub(), self.data, signature))

    def test_medium(self):
        """
        Check if medium security keys generate a valid signature.
        """
        key = self.ec.generate_key(u"medium")

        signature = key.signature(self.data)

        self.assertTrue(self.ec.is_valid_signature(key.pub(), self.data, signature))

    def test_high(self):
        """
        Check if high security keys generate a valid signature.
        """
        key = self.ec.generate_key(u"high")

        signature = key.signature(self.data)

        self.assertTrue(self.ec.is_valid_signature(key.pub(), self.data, signature))

    def test_curve25519(self):
        """
        Check if curve25519 keys generate a valid signature.
        """
        key = self.ec.generate_key(u"curve25519")

        signature = key.signature(self.data)

        self.assertTrue(self.ec.is_valid_signature(key.pub(), self.data, signature))

    def test_invalid_m2crypto(self):
        """
        Check if an M2Crypto key detects an invalid signature.
        """
        key = self.ec.generate_key(u"very-low")

        signature = ""

        self.assertFalse(self.ec.is_valid_signature(key.pub(), self.data, signature))

    def test_invalid_nacl(self):
        """
        Check if an libnacl key detects an invalid signature.
        """
        key = self.ec.generate_key(u"curve25519")

        signature = ""

        self.assertFalse(self.ec.is_valid_signature(key.pub(), self.data, signature))



from ...base import TestBase
from ...mocking.community import MockCommunity
from ...mocking.endpoint import AutoMockEndpoint, MockEndpointListener
from ....bootstrapping.dispersy.bootstrapper import DispersyBootstrapper


class TestDispersyBootstrapper(TestBase):

    def setUp(self):
        super().setUp()

        self.bootstrap_node = AutoMockEndpoint()
        self.bs_ep_listener = MockEndpointListener(self.bootstrap_node)
        self.bootstrap_node.open()

        self.bootstrapper = DispersyBootstrapper([self.bootstrap_node.wan_address], [], 60.0)
        self.overlay = MockCommunity()

    async def test_initialize(self):
        """
        Check if the special bootstrap addresses are added to the overlay's blacklist upon initialization.

        We don't test network DNS resolution here, which would contact the Internet.
        """
        result = self.bootstrapper.initialize(self.overlay)

        self.assertIn(self.bootstrap_node.wan_address, self.overlay.network.blacklist)
        self.assertTrue(result)

    async def test_get_addresses(self):
        """
        Check if the bootstrapper contacts the registered bootstrap nodes and doesn't return manual addresses.
        """
        addresses = await self.bootstrapper.get_addresses(self.overlay, 60.0)
        await self.deliver_messages()

        # There should be no manually walkable addresses.
        self.assertListEqual([], addresses)

        # However, the bootstrap node should have been contacted using the IPv8 protocol.
        self.assertEqual(1, len(self.bs_ep_listener.received_packets))

    async def test_get_addresses_timeout(self):
        """
        Check if a second call within the timeout gets dropped.
        """
        await self.bootstrapper.get_addresses(self.overlay, 60.0)
        await self.deliver_messages()

        await self.bootstrapper.get_addresses(self.overlay, 60.0)
        await self.deliver_messages()

        self.assertEqual(1, len(self.bs_ep_listener.received_packets))

    async def test_get_addresses_blacklist(self):
        """
        Check if the get_addresses ensures bootstrap nodes are in the blacklist.
        """
        self.overlay.network.blacklist.clear()

        await self.bootstrapper.get_addresses(self.overlay, 60.0)
        await self.deliver_messages()

        self.assertIn(self.bootstrap_node.wan_address, self.overlay.network.blacklist)

    async def test_keep_alive(self):
        """
        Check if the keep_alive tries to walk to a bootstrap node to keepe the connection alive.
        """
        self.bootstrapper.keep_alive(self.overlay)
        await self.deliver_messages()

        self.assertEqual(1, len(self.bs_ep_listener.received_packets))

    async def test_keep_alive_blacklist(self):
        """
        Check if the keep_alive adds walked nodes into the blacklist.
        """
        self.overlay.network.blacklist.clear()

        self.bootstrapper.keep_alive(self.overlay)
        await self.deliver_messages()

        self.assertIn(self.bootstrap_node.wan_address, self.overlay.network.blacklist)

    async def test_blacklist(self):
        """
        Check if the blacklist returns the added nodes in the blacklist.
        """
        self.assertListEqual([self.bootstrap_node.wan_address], self.bootstrapper.blacklist())



from asyncio import Future, all_tasks, sleep

from .base import TestBase
from ..requestcache import NumberCache, RandomNumberCache, RequestCache

CACHE_TIMEOUT = 0.01


class MockCache(RandomNumberCache):

    def __init__(self, request_cache):
        super(MockCache, self).__init__(request_cache, u"mock")
        self.timed_out = Future()

    @property
    def timeout_delay(self):
        return CACHE_TIMEOUT

    def on_timeout(self):
        self.timed_out.set_result(None)


class MockRegisteredCache(RandomNumberCache):

    def __init__(self, request_cache):
        super(MockRegisteredCache, self).__init__(request_cache, u"mock")
        self.timed_out = Future()
        self.register_future(self.timed_out)

    @property
    def timeout_delay(self):
        return CACHE_TIMEOUT

    def on_timeout(self):
        pass


class MockInfiniteCache(RandomNumberCache):

    def __init__(self, request_cache):
        super().__init__(request_cache, u"mock")
        self.timed_out = False

    @property
    def timeout_delay(self):
        return 8589934592

    def on_timeout(self):
        self.timed_out = True


class TestRequestCache(TestBase):

    def setUp(self):
        super().setUp()
        self.request_cache = RequestCache()

    async def test_shutdown(self):
        """
        Test if RequestCache does not allow new Caches after shutdown().
        """
        num_tasks = len(all_tasks())
        request_cache = RequestCache()  # This adds a task, don't use ``self.request_cache`` here!
        request_cache.add(MockCache(request_cache))
        self.assertEqual(len(all_tasks()), num_tasks + 2)
        await request_cache.shutdown()
        self.assertEqual(len(all_tasks()), num_tasks)
        request_cache.add(MockCache(request_cache))
        self.assertEqual(len(all_tasks()), num_tasks)

    async def test_timeout(self):
        """
        Test if the cache.on_timeout() is called after the cache.timeout_delay.
        """
        cache = MockCache(self.request_cache)
        self.request_cache.add(cache)
        await cache.timed_out

    async def test_add_duplicate(self):
        """
        Test if adding a cache twice returns None as the newly added cache.
        """
        cache = MockCache(self.request_cache)
        self.request_cache.add(cache)

        self.assertIsNone(self.request_cache.add(cache))

        await self.request_cache.shutdown()

    async def test_timeout_future_default_value(self):
        """
        Test if a registered future gets set to None on timeout.
        """
        cache = MockRegisteredCache(self.request_cache)
        self.request_cache.add(cache)
        self.assertEqual(None, (await cache.timed_out))
        await self.request_cache.shutdown()

    async def test_timeout_future_custom_value(self):
        """
        Test if a registered future gets set to a value on timeout.
        """
        cache = MockRegisteredCache(self.request_cache)
        self.request_cache.add(cache)

        cache.managed_futures[0] = (cache.managed_futures[0][0], 123)
        self.assertEqual(123, (await cache.timed_out))

        await self.request_cache.shutdown()

    async def test_timeout_future_exception(self):
        """
        Test if a registered future raises an exception on timeout.
        """
        cache = MockRegisteredCache(self.request_cache)
        self.request_cache.add(cache)

        cache.managed_futures[0] = (cache.managed_futures[0][0], RuntimeError())
        with self.assertRaises(RuntimeError):
            await cache.timed_out

        await self.request_cache.shutdown()

    async def test_cancel_future_after_shutdown(self):
        """
        Test if a registered future is cancelled when the RequestCache has shutdown.
        """
        await self.request_cache.shutdown()
        cache = MockRegisteredCache(self.request_cache)
        self.request_cache.add(cache)
        assert cache.managed_futures[0][0].done()

    async def test_cancel_future(self):
        """
        Test if a registered future gets canceled at shutdown.
        """
        cache = MockRegisteredCache(self.request_cache)
        self.request_cache.add(cache)
        await self.request_cache.shutdown()
        self.assertTrue(cache.timed_out.cancelled())

    async def test_passthrough_noargs(self):
        """
        Test if passthrough without arguments immediately times a cache out.
        """
        cache = MockInfiniteCache(self.request_cache)

        with self.request_cache.passthrough():
            self.request_cache.add(cache)
            await sleep(0.0)

        self.assertTrue(cache.timed_out)

    async def test_passthrough_timeout(self):
        """
        Test if passthrough respects the timeout value.
        """
        cache = MockInfiniteCache(self.request_cache)

        with self.request_cache.passthrough(timeout=10.0):
            self.request_cache.add(cache)
            await sleep(0.0)

        self.assertFalse(cache.timed_out)

    async def test_passthrough_filter_one_match(self):
        """
        Test if passthrough filters correctly with one filter, that matches
        """
        cache = MockInfiniteCache(self.request_cache)

        with self.request_cache.passthrough(MockInfiniteCache):
            self.request_cache.add(cache)
            await sleep(0.0)

        self.assertTrue(cache.timed_out)

    async def test_passthrough_filter_one_mismatch(self):
        """
        Test if passthrough filters correctly with one filter, that doesn't match
        """
        cache = MockInfiniteCache(self.request_cache)

        with self.request_cache.passthrough(MockRegisteredCache):
            self.request_cache.add(cache)
            await sleep(0.0)

        self.assertFalse(cache.timed_out)

    async def test_passthrough_filter_many_match(self):
        """
        Test if passthrough filters correctly with many filters, that all match
        """
        cache = MockInfiniteCache(self.request_cache)

        with self.request_cache.passthrough(MockInfiniteCache, RandomNumberCache, NumberCache):
            self.request_cache.add(cache)
            await sleep(0.0)

        self.assertTrue(cache.timed_out)

    async def test_passthrough_filter_some_match(self):
        """
        Test if passthrough filters correctly with many filters, for which some match
        """
        cache = MockInfiniteCache(self.request_cache)

        with self.request_cache.passthrough(MockRegisteredCache, MockCache, RandomNumberCache):
            self.request_cache.add(cache)
            await sleep(0.0)

        self.assertTrue(cache.timed_out)

    async def test_passthrough_filter_no_match(self):
        """
        Test if passthrough filters correctly with many filters, for which none match
        """
        cache = MockInfiniteCache(self.request_cache)

        with self.request_cache.passthrough(MockRegisteredCache, MockCache):
            self.request_cache.add(cache)
            await sleep(0.0)

        self.assertFalse(cache.timed_out)



from cryptography.hazmat.primitives.asymmetric.rsa import _modinv  # type:ignore


def format_polynomial(a, b, c):
    """
    Formats a polynomial cx^2 + bx + a into non-zero/non-one form.

    Ex. '0x^2 + 1x + 2' becomes 'x + 2'

    :param a: x^0 coefficient
    :param b: x^1 coefficient
    :param c: x^2 coefficient
    :returns: pretty format of polynomial
    """
    out = ''
    for (v, s) in [(a, ''), (b, 'x'), (c, 'x^2')]:
        if v:
            fmt_v = '' if abs(v) == 1 and s != '' else str(abs(v))
            out += (' + ' if out else '') + fmt_v + s
    if not out:
        out = '0'
    return out


class FP2Value(object):
    """
    Defines a rational value (a + bx + cx^2)/(aC + bCx + cCx^2)(mod 1 + x + x^2, mod p).
    """

    def __init__(self, mod, a=0, b=0, c=0, aC=1, bC=0, cC=0):
        """
        Intialize a value mod 'mod' of two quadratic polynomials divided by each other modulo (x^2 + x + 1).

        :param mod: the modulus
        :param a: the coefficient of 1
        :param b: the coefficient of x
        :param c: the coefficient of x^2
        :param aC: (the coefficient of 1) ^ -1
        :param bC: (the coefficient of x) ^ -1
        :param cC: (the coefficient of x) ^ -2
        """
        self.mod = mod
        self.a, self.b, self.c, self.aC, self.bC, self.cC = a % mod, b % mod, c % mod, aC % mod, bC % mod, cC % mod

    def __str__(self):
        """
        Format this value as a string.
        """
        numerator = format_polynomial(self.a, self.b, self.c)
        denominator = format_polynomial(self.aC, self.bC, self.cC)
        if denominator == '1':
            return numerator
        else:
            return "(%s)/(%s)" % (numerator, denominator)

    def __add__(self, other):
        """
        Add this value to another value and return a new FP2Value.
        """
        assert self.mod == other.mod

        a = (self.aC * other.a - self.cC * other.a + self.a * other.aC - self.c * other.aC - self.bC * other.b
             + self.cC * other.b - self.aC * other.c + self.bC * other.c - self.a * other.cC + self.b * other.cC)
        b = (self.bC * other.a - self.cC * other.a + self.b * other.aC - self.c * other.aC + self.aC * other.b
             - self.bC * other.b + self.a * other.bC - self.b * other.bC - self.aC * other.c + self.cC * other.c
             - self.a * other.cC + self.c * other.cC)
        aC = (self.aC * other.aC - self.cC * other.aC - self.bC * other.bC
              + self.cC * other.bC - self.aC * other.cC + self.bC * other.cC)
        bC = (self.bC * other.aC - self.cC * other.aC + self.aC * other.bC
              - self.bC * other.bC - self.aC * other.cC + self.cC * other.cC)
        return FP2Value(self.mod, a=a, b=b, aC=aC, bC=bC)

    def __sub__(self, other):
        """
        Subtract another value from this value and return a new FP2Value.
        """
        assert self.mod == other.mod
        a = (-self.aC * other.a + self.cC * other.a + self.a * other.aC - self.c * other.aC + self.bC * other.b
             - self.cC * other.b - self.b * other.bC + self.c * other.bC + self.aC * other.c - self.bC * other.c
             - self.a * other.cC + self.b * other.cC)
        b = (-self.bC * other.a + self.cC * other.a + self.b * other.aC - self.c * other.aC - self.aC * other.b
             + self.bC * other.b + self.a * other.bC - self.b * other.bC + self.aC * other.c - self.cC * other.c
             - self.a * other.cC + self.c * other.cC)
        aC = (self.aC * other.aC - self.cC * other.aC - self.bC * other.bC
              + self.cC * other.bC - self.aC * other.cC + self.bC * other.cC)
        bC = (self.bC * other.aC - self.cC * other.aC + self.aC * other.bC
              - self.bC * other.bC - self.aC * other.cC + self.cC * other.cC)
        return FP2Value(self.mod, a=a, b=b, aC=aC, bC=bC)

    def __mul__(self, other):
        """
        Multiply this value with another value and return a new FP2Value.
        """
        assert self.mod == other.mod
        a = (self.a * other.a - self.c * other.a - self.b * other.b
             + self.c * other.b - self.a * other.c + self.b * other.c)
        b = (self.b * other.a - self.c * other.a + self.a * other.b
             - self.b * other.b - self.a * other.c + self.c * other.c)
        aC = (self.aC * other.aC - self.cC * other.aC - self.bC * other.bC
              + self.cC * other.bC - self.aC * other.cC + self.bC * other.cC)
        bC = (self.bC * other.aC - self.cC * other.aC + self.aC * other.bC
              - self.bC * other.bC - self.aC * other.cC + self.cC * other.cC)
        return FP2Value(self.mod, a=a, b=b, aC=aC, bC=bC)

    def __floordiv__(self, other):
        """
        Divide this value by another value and return a new FP2Value.
        """
        assert self.mod == other.mod
        a = (self.a * other.aC - self.c * other.aC - self.b * other.bC
             + self.c * other.bC - self.a * other.cC + self.b * other.cC)
        b = (self.b * other.aC - self.c * other.aC + self.a * other.bC
             - self.b * other.bC - self.a * other.cC + self.c * other.cC)
        aC = (self.aC * other.a - self.cC * other.a - self.bC * other.b
              + self.cC * other.b - self.aC * other.c + self.bC * other.c)
        bC = (self.bC * other.a - self.cC * other.a + self.aC * other.b
              - self.bC * other.b - self.aC * other.c + self.cC * other.c)
        return FP2Value(self.mod, a=a, b=b, aC=aC, bC=bC)

    def __eq__(self, other):
        """
        Check equality with another value.
        """
        if not isinstance(other, FP2Value):
            return False
        divd = (self // other).normalize()
        return all([divd.a == divd.aC, divd.b == divd.bC, divd.c == divd.cC])

    def __hash__(self):
        """
        Equality is not trivial. We hash everything to 0 for the full equality check.
        """
        return 0

    def intpow(self, power):
        """
        Raise this value by a given power (int).

        :param power: the power to raise this value by
        :type power: int
        """
        n = -power if power < 0 else power
        R = FP2Value(self.mod, 1)
        U = self
        while n > 0:
            if (n % 2) == 1:
                R *= U
            U *= U
            n = n // 2
        return R.inverse().normalize() if power < 0 else R

    def normalize(self):
        """
        Normalize to aC = 1: this is the best human-readable form.

        Ex. '20/4' becomes '5'
            '4 + 4x/2 + 2x' becomes '2 + 2x/1 + x'
        """
        mp = _modinv(self.aC % self.mod, self.mod)
        if mp > 0:
            a = (self.a * mp) % self.mod
            b = (self.b * mp) % self.mod
            c = (self.c * mp) % self.mod
            aC = 1
            bC = (self.bC * mp) % self.mod
            cC = (self.cC * mp) % self.mod
            return FP2Value(self.mod, a, b, c, aC, bC, cC)
        return FP2Value(self.mod, self.a, self.b, self.c, self.aC, self.bC, self.cC)

    def inverse(self):
        """
        Return the inverse of this value.
        """
        return FP2Value(self.mod, a=self.aC, b=self.bC, c=self.cC, aC=self.a, bC=self.b, cC=self.c)

    def wp_nominator(self):
        """
        Return the '1' and 'x' coefficients as a new value.
        """
        return FP2Value(self.mod, self.a, self.b)

    def wp_denom_inverse(self):
        """
        Return the '1^-1' and 'x^-1' coefficients modular inverse.
        """
        iq = FP2Value(self.mod, self.aC * self.aC - self.aC * self.bC + self.bC * self.bC)
        a = FP2Value(self.mod, self.aC - self.bC) // iq
        b = FP2Value(self.mod, -self.bC) // iq
        return FP2Value(self.mod, a.normalize().a, b.normalize().a)

    def wp_compress(self):
        """
        Compress this FP2 value into an FP2Value only containing `a` and `b` values.

        This is an expensive operation.
        """
        assert self.c == 0
        assert self.cC == 0
        normalized = self.normalize()
        return normalized.wp_nominator() * normalized.wp_denom_inverse()



import os
from functools import reduce

from ..base import TestBase
from ..mocking.community import MockCommunity
from ...bootstrapping.dispersy.bootstrapper import DispersyBootstrapper
from ...messaging.payload_headers import BinMemberAuthenticationPayload, GlobalTimeDistributionPayload
from ...peerdiscovery.payload import DiscoveryIntroductionRequestPayload


class TestDiscoveryCommunity(TestBase):

    def setUp(self):
        super(TestDiscoveryCommunity, self).setUp()
        self.tracker = MockCommunity()

        node_count = 2
        self.overlays = [MockCommunity() for _ in range(node_count)]
        for overlay in self.overlays:
            overlay.network.blacklist.append(self.tracker.endpoint.wan_address)
            bootstrapper = DispersyBootstrapper([self.tracker.endpoint.wan_address], [], bootstrap_timeout=0.0)
            overlay.bootstrappers = [bootstrapper]

    async def tearDown(self):
        await self.tracker.unload()
        for overlay in self.overlays:
            await overlay.unload()
        return await super(TestDiscoveryCommunity, self).tearDown()

    async def test_deprecated_introduction(self):
        """
        Check if we can handle the deprecated Discovery introduction request as a normal one.
        """
        global_time = self.overlays[0].claim_global_time()
        payload = DiscoveryIntroductionRequestPayload(b"a" * 20,
                                                      self.overlays[1].endpoint.wan_address,
                                                      self.overlays[0].my_estimated_lan,
                                                      self.overlays[0].my_estimated_wan,
                                                      True,
                                                      u"unknown",
                                                      global_time,
                                                      b'')
        auth = BinMemberAuthenticationPayload(self.overlays[0].my_peer.public_key.key_to_bin())
        dist = GlobalTimeDistributionPayload(global_time)

        packet = self.overlays[0]._ez_pack(self.overlays[0]._prefix, 246, [auth, dist, payload])
        self.overlays[1].on_old_introduction_request(self.overlays[0].endpoint.wan_address, packet)

        await self.deliver_messages()

        self.assertEqual(1, len(self.overlays[1].network.verified_peers))

    async def test_bootstrap(self):
        """
        Check if we can bootstrap our peerdiscovery.
        """
        # Both other overlays contact the tracker
        self.overlays[0].bootstrap()
        self.overlays[1].bootstrap()
        await self.deliver_messages()

        self.assertEqual(len(self.tracker.network.verified_peers), 2)

        # Now that the tracker knows both others, they should be introduced to each other
        self.overlays[0].bootstrap()
        self.overlays[1].bootstrap()
        await self.deliver_messages()

        for overlay in self.overlays:
            intros = overlay.network.get_introductions_from(self.tracker.my_peer)
            # Over time we get more than one option per network interface.
            # Usually deliver_messages will not deliver more than 1 option, but we could get more.
            self.assertGreaterEqual(len(intros), 1)
            self.assertNotIn(overlay.my_peer.mid, intros)
            self.assertNotIn(self.tracker.my_peer.mid, intros)

    async def test_cross_peer(self):
        """
        If we have different peers under our control, don't claim to be the other identity.
        """
        custom_community_id = os.urandom(20)

        class OtherMockCommunity(MockCommunity):
            community_id = custom_community_id
        custom_overlay = OtherMockCommunity()
        custom_overlay.my_peer.address = self.overlays[0].my_peer.address
        self.overlays.append(custom_overlay)
        self.overlays[0].network.register_service_provider(custom_community_id, custom_overlay)

        self.overlays[0].walk_to(self.overlays[1].my_peer.address)
        await self.deliver_messages()

        discovered = reduce(lambda a, b: a | b, self.overlays[1].network.services_per_peer.values(), set())

        self.assertEqual(len(self.overlays[1].network.services_per_peer), 2)
        self.assertSetEqual(discovered, {MockCommunity.community_id, custom_community_id})



import struct

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.hkdf import HKDFExpand

import libnacl
from libnacl.aead import AEAD

from ...keyvault.crypto import ECCrypto, LibNaCLPK


class CryptoException(Exception):
    pass


class TunnelCrypto(ECCrypto):

    def initialize(self, key):
        self.key = key
        assert isinstance(self.key, LibNaCLPK), type(self.key)

    def is_key_compatible(self, key):
        return isinstance(key, LibNaCLPK)

    def generate_diffie_secret(self):
        tmp_key = self.generate_key("curve25519")
        X = tmp_key.key.pk

        return tmp_key, X

    def generate_diffie_shared_secret(self, dh_received, key=None):
        if key is None:
            key = self.key

        tmp_key = self.generate_key("curve25519")
        y = tmp_key.key.sk
        Y = tmp_key.key.pk
        shared_secret = libnacl.crypto_box_beforenm(dh_received, y) + libnacl.crypto_box_beforenm(dh_received, key.key.sk)

        AUTH = libnacl.crypto_auth(Y, shared_secret[:32])
        return shared_secret, Y, AUTH

    def verify_and_generate_shared_secret(self, dh_secret, dh_received, auth, B):
        shared_secret = libnacl.crypto_box_beforenm(dh_received, dh_secret.key.sk) + libnacl.crypto_box_beforenm(B, dh_secret.key.sk)
        libnacl.crypto_auth_verify(auth, dh_received, shared_secret[:32])

        return shared_secret

    def generate_session_keys(self, shared_secret):
        hkdf = HKDFExpand(algorithm=hashes.SHA256(), backend=default_backend(), length=72, info=b"key_generation")
        key = hkdf.derive(shared_secret)

        kf = key[:32]
        kb = key[32:64]
        sf = key[64:68]
        sb = key[68:72]
        return [kf, kb, sf, sb, 1, 1]

    def get_session_keys(self, keys, direction):
        # increment salt_explicit
        keys[direction + 4] += 1
        return keys[direction], keys[direction + 2], keys[direction + 4]

    def encrypt_str(self, content, key, salt, salt_explicit):
        # return the encrypted content prepended with salt_explicit
        aead = AEAD(key)
        _, _, ciphertext = aead.encrypt(content, b'',
                                        nonce=salt + struct.pack('!q', salt_explicit),
                                        pack_nonce_aad=False)
        return struct.pack('!q', salt_explicit) + ciphertext

    def decrypt_str(self, content, key, salt):
        # content contains the tag and salt_explicit in plaintext
        if len(content) < 24:
            raise CryptoException("truncated content")

        aead = AEAD(key)
        return aead.decrypt(salt + content, 0)



from .gabi.proofs import ProofD, createChallenge
from ..primitives.structs import ipack, iunpack


def serialize_proof_d(proof_d):
    return (ipack(proof_d.C)
            + ipack(proof_d.A)
            + ipack(proof_d.EResponse)
            + ipack(proof_d.VResponse)
            + ipack(proof_d.AResponses[0]))


def unserialize_proof_d(s):
    C, rem = iunpack(s)
    A, rem = iunpack(rem)
    EResponse, rem = iunpack(rem)
    VResponse, rem = iunpack(rem)
    fa_response, rem = iunpack(rem)
    return ProofD(C, A, EResponse, VResponse, {0: fa_response}, {})


def challenge_response(my_proof_d, Z, challenge):
    challenge_verif, _ = iunpack(challenge)
    return ipack(createChallenge(challenge_verif, challenge_verif, [my_proof_d.A, Z], False))



from asyncio import sleep
from base64 import b64encode

from ..REST.rest_base import RESTTestBase, partial_cls
from ...attestation.identity.community import IdentityCommunity
from ...attestation.identity.manager import IdentityManager
from ...attestation.wallet.community import AttestationCommunity


class TestAttestationEndpoint(RESTTestBase):
    """
    Class for testing the REST API of the AttestationEndpoint
    """

    async def setUp(self):
        super(TestAttestationEndpoint, self).setUp()
        identity_manager = IdentityManager(u":memory:")
        await self.initialize([partial_cls(AttestationCommunity, working_directory=':memory:'),
                               partial_cls(IdentityCommunity, identity_manager=identity_manager)], 2)

    async def make_outstanding(self, node):
        """
        Forward a request for outstanding attestation requests.
        """
        return await self.make_request(node, 'attestation', 'get', {'type': 'outstanding'})

    async def make_verification_output(self, node):
        """
        Forward a request for the verification outputs.
        """
        return await self.make_request(node, 'attestation', 'get', {'type': 'verification_output'})

    async def make_peers(self, node):
        """
        Forward a request for the known peers in the network.
        """
        return await self.make_request(node, 'attestation', 'get', {'type': 'peers'})

    async def make_attributes(self, node):
        """
        Forward a request for the attributes of a peer.
        """
        return await self.make_request(node, 'attestation', 'get', {'type': 'attributes'})

    async def wait_for_attributes(self, node):
        """
        Forward a request for the attributes of a peer.
        """
        attributes = await self.make_request(node, 'attestation', 'get', {'type': 'attributes'})
        while not attributes:
            attributes = await self.make_request(node, 'attestation', 'get', {'type': 'attributes'})
            await self.deliver_messages()
        return attributes

    async def make_dht_block(self, node, public_key):
        """
        Forward a request for the latest TC block of a peer
        """
        return await self.make_request(node, 'attestation', 'get', {'public_key': public_key})

    async def make_drop_identity(self, node):
        """
        Forward a request for dropping a peer's identity.
        """
        return await self.make_request(node, 'attestation', 'get', {'type': 'drop_identity'})

    async def make_outstanding_verify(self, node):
        """
        Forward a request which requests information on the outstanding verify requests
        """
        return await self.make_request(node, 'attestation', 'get', {'type': 'outstanding_verify'})

    async def make_attestation_request(self, node, attribute_name, mid, metadata=None):
        """
        Forward a request for the attestation of an attribute.
        """

        # Add the type of the request (request), and the rest of the parameters
        request_parameters = {'type': 'request',
                              'id_format': 'id_metadata',
                              'attribute_name': attribute_name,
                              'mid': mid}
        if metadata:
            request_parameters['metadata'] = metadata
        return await self.make_request(node, 'attestation', 'post', request_parameters)

    async def make_attest(self, node, attribute_name, attribute_value, mid):
        """
        Forward a request which attests an attestation request.
        """
        return await self.make_request(node, 'attestation', 'post', {'type': 'attest',
                                                                     'attribute_name': attribute_name,
                                                                     'attribute_value': attribute_value,
                                                                     'mid': mid})

    async def make_verify(self, node, attribute_hash, attribute_values, mid):
        """
        Forward a request which demands the verification of an attestation
        """
        return await self.make_request(node, 'attestation', 'post', {'type': 'verify',
                                                                     'attribute_hash': attribute_hash,
                                                                     'mid': mid,
                                                                     'attribute_values': attribute_values})

    async def make_allow_verify(self, node, attribute_name, mid):
        """
        Forward a request which requests that verifications be allowed for a particular peer for a particular attribute
        """
        return await self.make_request(node, 'attestation', 'post', {'type': 'allow_verify',
                                                                     'attribute_name': attribute_name,
                                                                     'mid': mid})

    async def create_attestation_request(self, node, attribute_name, metadata=None):
        peer_list = await self.wait_for_peers(node)
        for mid in peer_list:
            await self.make_attestation_request(node, attribute_name, mid, metadata=metadata)

    async def wait_for_peers(self, node):
        """
        Wait until this peer receives a non-empty list of fellow peers in the network
        """
        peer_list = await self.make_peers(node)
        while not peer_list:
            await sleep(.1)
            peer_list = await self.make_peers(node)
        return peer_list

    async def wait_for_outstanding_requests(self, node):
        """
        Wait until this peer receives a non-empty list of outstanding attestation requests
        """
        outstanding_requests = await self.make_outstanding(node)
        while not outstanding_requests:
            await sleep(.1)
            outstanding_requests = await self.make_outstanding(node)
        return [(x[0].encode('utf-8'), x[1], x[2]) for x in outstanding_requests]

    async def attest_all_outstanding_requests(self, node, attribute_name, attribute_value):
        """
        Forward an attestation for each of the outstanding attestation requests

        :param param_dict: the parameters required to contact a well-known peer for the POST and GET requests
        :return: a list of the outstanding requests and their (empty if successful) request responses
        """

        outstanding_requests = await self.wait_for_outstanding_requests(node)
        self.assertFalse(outstanding_requests == [], "Something went wrong, no request was received.")

        # Collect the responses of the attestations; if functioning properly, this should be a list of empty strings
        responses = []

        for outstanding_request in outstanding_requests:
            # The attestation value is already computed, so don't bother recomputing it here
            mid = outstanding_request[0].decode('utf-8')
            attribute_value = b64encode(attribute_value).decode('utf-8')\
                if isinstance(attribute_value, bytes) else attribute_value
            response = await self.make_attest(node, attribute_name, attribute_value, mid)
            responses.append(response)

        return outstanding_requests, responses

    async def verify_all_attestations(self, node, peer_mids, attribute_hash, attribute_values):
        """
        Forward an attestation verification for a set of attestations

        :param peer_mids: the set of peer mids to which a verification request will be sent
        :param param_dict: the parameters required to contact a well-known peer for the POST: verify request
        :return: the verification responses, as returned by the well-known peer. Ideally these should be all empty
        """
        assert peer_mids, "Attestation list is empty"

        verification_responses = []

        for mid in peer_mids:
            mid = b64encode(mid).decode('utf-8') if isinstance(mid, bytes) else mid
            intermediary_response = await self.make_verify(node, attribute_hash, attribute_values, mid)
            verification_responses.append(intermediary_response)

        return verification_responses

    async def test_get_peers_request(self):
        """
        Test the (GET: peers request) type
        """

        await self.introduce_nodes()
        other_peer_mids = [b64encode(self.mid(1)).decode('utf-8')]
        result = await self.wait_for_peers(self.node(0))
        self.assertTrue(any(x in other_peer_mids for x in result), "Could not find the second peer.")

    async def test_get_outstanding_requests(self):
        """
        Test the (GET: outstanding) request type
        """

        await self.introduce_nodes()
        await self.create_attestation_request(self.node(1), 'QR')

        result = await self.wait_for_outstanding_requests(self.node(0))

        mid = b64encode(self.mid(1))
        self.assertTrue(any((x[0] == mid and x[1] == 'QR' for x in result)),
                        "Could not find the outstanding request forwarded by the second peer")

    async def test_get_verification_output(self):
        """
        Test the (GET: verification output) request type
        """

        # Forward the attestations to the well-known peer
        await self.introduce_nodes()
        await self.create_attestation_request(self.node(1), 'QR')
        await self.attest_all_outstanding_requests(self.node(0), 'QR', b'data')

        # Get the hash of the attestation to be validated (the one which was just attested)
        attributes = await self.wait_for_attributes(self.node(1))
        attribute_hash = attributes[0][1]

        # Forward the actual verification
        verification_responses = await self.verify_all_attestations(self.node(0),
                                                                    [self.mid(1)],
                                                                    attribute_hash, 'YXNk,YXNkMg==')
        self.assertTrue(all("success" in x and x["success"] for x in verification_responses),
                        "At least one of the verification responses was non-empty.")

        # Unlock the verification
        outstanding_verifications = []
        while not outstanding_verifications:
            outstanding_verifications = await self.make_outstanding_verify(self.node(1))
            self.assertIsNotNone(outstanding_verifications, "Could not retrieve any outstanding verifications")
            await self.deliver_messages()

        mid = outstanding_verifications[0][0]
        await self.make_allow_verify(self.node(1), 'QR', mid)
        await sleep(.1)

        # Get the output
        verification_output = await self.make_verification_output(self.node(0))
        self.assertTrue([["YXNk", 0.0], ["YXNkMg==", 0.0]] in verification_output.values(),
                        "Something went wrong with the verification. Unexpected output values.")

    async def test_get_outstanding_verify(self):
        """
        Test the (GET: outstanding verify) request type
        """

        # Forward the attestations to the well-known peer
        await self.introduce_nodes()
        await self.create_attestation_request(self.node(1), 'QR')
        await self.attest_all_outstanding_requests(self.node(0), 'QR', b'data')

        # Get the hash of the attestation to be validated (the one which was just attested)
        attributes = await self.wait_for_attributes(self.node(1))
        attribute_hash = attributes[0][1]

        # Forward the actual verification
        verification_responses = await self.verify_all_attestations(self.node(0),
                                                                    [self.mid(1)],
                                                                    attribute_hash, 'YXNk,YXNkMg==')
        self.assertTrue(all("success" in x and x["success"] for x in verification_responses),
                        "At least one of the verification responses was non-empty.")

        # Unlock the verification
        outstanding_verifications = []
        while not outstanding_verifications:
            outstanding_verifications = await self.make_outstanding_verify(self.node(1))
            self.assertIsNotNone(outstanding_verifications, "Could not retrieve any outstanding verifications")
            await self.deliver_messages()

        # Retrieve only the mids
        result = [x[0] for x in outstanding_verifications]
        self.assertTrue(any(x in result for x in [b64encode(self.mid(0)).decode('utf-8')]),
                        "Something went wrong. Could not find a master peer mid in the "
                        "outstanding verification requests.")

    async def test_get_attributes(self):
        """
        Test the (GET: attributes) request type
        """
        await self.introduce_nodes()
        await self.create_attestation_request(self.node(1), 'QR')
        await self.attest_all_outstanding_requests(self.node(0), 'QR', 'data')

        # Get the hash of the attestation to be validated (the one which was just attested)
        attributes = await self.wait_for_attributes(self.node(1))
        self.assertTrue(attributes[0][0] == 'QR' and attributes[0][1] != "",
                        "The response was not as expected. This would suggest that something went wrong with "
                        "the attributes request.")

    async def test_get_drop_identity(self):
        """
        Test the (GET: drop identity) request type
        """

        await self.introduce_nodes()
        await self.create_attestation_request(self.node(1), 'QR')
        outstanding_requests = await self.wait_for_outstanding_requests(self.node(0))
        self.assertFalse(outstanding_requests == [], "The attestation requests were not received.")

        # Ensure that no block/attestation exists
        attributes = await self.make_attributes(self.node(1))
        self.assertEqual(attributes, [], "Something's wrong, there shouldn't be any blocks.")

        # Attest the outstanding request. This should mean that the attribute DB is non-empty in the well-known peer
        await self.attest_all_outstanding_requests(self.node(0), 'QR', 'data')

        # Ensure that the attestation has been completed
        attributes = await self.wait_for_attributes(self.node(1))
        self.assertNotEqual(attributes, [], "Something's wrong, the attribute list should be non-empty.")

        # Drop the identity
        result = await self.make_drop_identity(self.node(0))
        self.assertIn("success", result, "The identity could not be dropped. Success parameter not in response.")
        self.assertTrue(result["success"], "The identity could not be dropped, not successful.")

        # Make sure the identity was successfully dropped
        result = await self.make_attributes(self.node(0))
        self.assertEqual(result, [], 'The identity could not be dropped. Block DB still populated.')

        result = await self.make_outstanding(self.node(0))
        self.assertEqual(result, [], 'The identity could not be dropped. Outstanding requests still remaining.')

    async def test_post_attestation_request(self):
        """
        Test the (POST: request) request type
        """

        # This should return an empty response
        outstanding_requests = await self.make_outstanding(self.node(0))
        self.assertEqual(outstanding_requests, [], "Something went wrong, there should be no outstanding requests.")

        await self.introduce_nodes()
        await self.create_attestation_request(self.node(1), 'QR')

        # This should return a non-empty response
        outstanding_requests = await self.wait_for_outstanding_requests(self.node(0))
        self.assertFalse(outstanding_requests == [], "Something went wrong, no request was received.")

    async def test_post_attest(self):
        """
        Test the (POST: attest) request type
        """

        await self.introduce_nodes()
        await self.create_attestation_request(self.node(1), 'QR')

        attributes = await self.make_attributes(self.node(1))
        self.assertTrue(len(attributes) == 0, "There mustn't already be any attestations in the other peer.")

        responses = await self.attest_all_outstanding_requests(self.node(0), 'QR', 'data')
        request_responses = [response for response in responses[1]]
        self.assertTrue(all("success" in x and x["success"] for x in request_responses),
                        "Something went wrong, not all responses were successful.")

        attributes = await self.wait_for_attributes(self.node(1))
        self.assertTrue(len(attributes) == 1, "There should only be one attestation in the DB.")
        self.assertTrue(attributes[0][0] == 'QR', "Expected attestation for %s, got it for "
                                                  "%s" % ('QR', attributes[0][0]))

        attributes = await self.make_attributes(self.node(0))
        self.assertTrue(len(attributes) == 0, "There should be no attribute in the DB of the attester.")

    async def test_post_verify(self):
        """
        Test the (POST: verify) request type
        """

        # Forward the attestations to the well-known peer
        await self.introduce_nodes()
        await self.create_attestation_request(self.node(1), 'QR')
        await self.attest_all_outstanding_requests(self.node(0), 'QR', 'data')

        # Get the hash of the attestation to be validated (the one which was just attested)
        attributes = await self.wait_for_attributes(self.node(1))
        attribute_hash = attributes[0][1]

        # Forward the actual verification
        verification_responses = await self.verify_all_attestations(self.node(0),
                                                                    [self.mid(1)],
                                                                    attribute_hash, 'YXNk,YXNkMg==')
        self.assertTrue(all("success" in x and x["success"] for x in verification_responses),
                        "At least one of the verification responses was non-empty.")

    async def test_post_allow_verify(self):
        """
        Test the (POST: allow verify) request type
        """

        # Forward the attestations to the well-known peer
        await self.introduce_nodes()
        await self.create_attestation_request(self.node(1), 'QR')
        await self.attest_all_outstanding_requests(self.node(0), 'QR', 'data')

        # Get the hash of the attestation to be validated (the one which was just attested)
        attributes = await self.wait_for_attributes(self.node(1))
        attribute_hash = attributes[0][1]

        # Forward the actual verification
        verification_responses = await self.verify_all_attestations(self.node(0),
                                                                    [self.mid(1)],
                                                                    attribute_hash, 'YXNk,YXNkMg==')
        self.assertTrue(all("success" in x and x["success"] for x in verification_responses),
                        "At least one of the verification responses was non-empty.")

        # Unlock the verification
        outstanding_verifications = []
        while not outstanding_verifications:
            outstanding_verifications = await self.make_outstanding_verify(self.node(1))
            self.assertIsNotNone(outstanding_verifications, "Could not retrieve any outstanding verifications")
            await self.deliver_messages()

        mid = outstanding_verifications[0][0]
        response = await self.make_allow_verify(self.node(1), 'QR', mid)
        self.assertIn("success", response, "The attestion could not be unlocked: success not in JSON response")
        self.assertTrue(response["success"], "The attestation could not be unlocked: not successful.")



import binascii

from ..REST.rest_base import RESTTestBase
from ..mocking.community import MockCommunity
from ...keyvault.crypto import default_eccrypto
from ...messaging.interfaces.statistics_endpoint import StatisticsEndpoint
from ...peer import Peer


def hexlify(value):
    return binascii.hexlify(value).decode()


class MockCommunity2(MockCommunity):
    community_id = b'DifferentCommunityID'


class TestOverlaysEndpoint(RESTTestBase):

    async def setUp(self):  # pylint: disable=W0236
        super().setUp()
        await self.initialize([], 1)
        self.ipv8 = self.node(0)

    def mount_statistics(self, i, add_mock_community=True):
        self.node(i).endpoint = StatisticsEndpoint(self.node(i).endpoint)
        self.node(i).rest_manager.root_endpoint.endpoints['/overlays'].statistics_supported = True

        if add_mock_community:
            self.node(i).overlay = self.add_mock_community(i)

    def add_mock_community(self, i, overlay_class=MockCommunity):
        mock_community = overlay_class()
        mock_community.endpoint = self.node(i).endpoint
        self.node(i).overlays.append(mock_community)
        return mock_community

    async def test_no_overlays(self):
        """
        Check if the overlays endpoint returns no overlays if it has no overlays.
        """
        response = await self.make_request(self.ipv8, "overlays", "GET")

        self.assertIn("overlays", response)
        self.assertListEqual([], response["overlays"])

    async def test_one_overlay_no_peers(self):
        """
        Check if the overlays endpoint returns one overlay if it has one overlay.
        """
        mock_community = MockCommunity()
        expected_id = hexlify(mock_community.community_id)
        expected_peer = hexlify(mock_community.my_peer.public_key.key_to_bin())
        mock_community.update_global_time(1337)
        self.ipv8.overlays.append(mock_community)

        response = await self.make_request(self.ipv8, "overlays", "GET")

        self.assertIn("overlays", response)
        self.assertEqual(1, len(response["overlays"]))
        self.assertEqual(expected_id, response["overlays"][0]["id"])
        self.assertEqual(expected_peer, response["overlays"][0]["my_peer"])
        self.assertEqual(mock_community.global_time, response["overlays"][0]["global_time"])
        self.assertListEqual([], response["overlays"][0]["peers"])
        self.assertEqual("MockCommunity", response["overlays"][0]["overlay_name"])
        self.assertDictEqual({}, response["overlays"][0]["statistics"])

    async def test_one_overlay_one_peer(self):
        """
        Check if the overlays endpoint correctly returns its one peer for its one overlay.
        """
        mock_community = MockCommunity()
        expected_peer = Peer(default_eccrypto.generate_key("very-low"), ("1.2.3.4", 5))
        mock_community.network.add_verified_peer(expected_peer)
        mock_community.network.discover_services(expected_peer, [mock_community.community_id])
        self.ipv8.overlays.append(mock_community)

        response = await self.make_request(self.ipv8, "overlays", "GET")

        self.assertIn("overlays", response)
        self.assertEqual(1, len(response["overlays"]))
        self.assertEqual(1, len(response["overlays"][0]["peers"]))
        self.assertEqual("1.2.3.4", response["overlays"][0]["peers"][0]["ip"])
        self.assertEqual(5, response["overlays"][0]["peers"][0]["port"])
        self.assertEqual(hexlify(expected_peer.public_key.key_to_bin()),
                         response["overlays"][0]["peers"][0]["public_key"])

    async def test_one_overlay_multiple_peers(self):
        """
        Check if the overlays endpoint correctly returns its peers for its one overlay.
        """
        mock_community = MockCommunity()
        peer_count = 3
        for _ in range(peer_count):
            expected_peer = Peer(default_eccrypto.generate_key("very-low"), ("1.2.3.4", 5))
            mock_community.network.add_verified_peer(expected_peer)
            mock_community.network.discover_services(expected_peer, [mock_community.community_id])
        self.ipv8.overlays.append(mock_community)

        response = await self.make_request(self.ipv8, "overlays", "GET")

        self.assertIn("overlays", response)
        self.assertEqual(1, len(response["overlays"]))
        self.assertEqual(peer_count, len(response["overlays"][0]["peers"]))

        # peer_count peers with the same address and distinct keys:
        known_keys = set()
        for i in range(peer_count):
            self.assertEqual("1.2.3.4", response["overlays"][0]["peers"][i]["ip"])
            self.assertEqual(5, response["overlays"][0]["peers"][i]["port"])
            self.assertNotIn(response["overlays"][0]["peers"][i]["public_key"], known_keys)
            known_keys.add(response["overlays"][0]["peers"][i]["public_key"])

    async def test_one_overlay_statistics(self):
        """
        Check if the overlays endpoint returns overlay statistics correctly for one overlay.
        """
        self.mount_statistics(0)
        self.node(0).endpoint.add_sent_stat(self.overlay(0).get_prefix(), 245, 1337)

        expected_stats = {
            'bytes_down': 0,
            'bytes_up': 1337,
            'diff_time': 0.0,
            'num_down': 0,
            'num_up': 1
        }

        response = await self.make_request(self.ipv8, "overlays", "GET")

        self.assertIn("overlays", response)
        self.assertEqual(1, len(response["overlays"]))
        self.assertDictEqual(expected_stats, response["overlays"][0]["statistics"])

    async def test_multiple_overlays(self):
        """
        Check if the overlays endpoint returns multiple overlays.
        """
        expected_id = hexlify(MockCommunity.community_id)
        overlay_count = 3
        for i in range(overlay_count):
            mock_community = MockCommunity()
            mock_community.update_global_time(100 * i)
            self.ipv8.overlays.append(mock_community)

        response = await self.make_request(self.ipv8, "overlays", "GET")

        self.assertIn("overlays", response)
        self.assertEqual(3, len(response["overlays"]))

        # overlay_count overlays with the same community_id and distinct global_time:
        known_global_times = set()
        for i in range(overlay_count):
            self.assertEqual(expected_id, response["overlays"][i]["id"])
            self.assertNotIn(response["overlays"][i]["global_time"], known_global_times)
            known_global_times.add(response["overlays"][i]["global_time"])

    async def test_statistics_no_overlays(self):
        """
        Check if no statistics are returned if no overlays are loaded.
        """
        response = await self.make_request(self.ipv8, "overlays/statistics", "GET")

        self.assertIn("statistics", response)
        self.assertListEqual([], response["statistics"])

    async def test_statistics_one_overlay(self):
        """
        Check if statistics are returned for one loaded overlay.
        """
        self.mount_statistics(0)
        self.node(0).endpoint.add_sent_stat(self.overlay(0).get_prefix(), 245, 1337, 42.0)

        expected_stats = {
            'identifier': 245,
            'bytes_down': 0,
            'bytes_up': 1337,
            'num_down': 0,
            'num_up': 1,
            'first_measured_up': 42.0,
            'first_measured_down': 0,
            'last_measured_up': 42.0,
            'last_measured_down': 0
        }

        response = await self.make_request(self.ipv8, "overlays/statistics", "GET")

        self.assertIn("statistics", response)
        self.assertEqual(1, len(response["statistics"]))
        self.assertIn("MockCommunity", response["statistics"][0])
        self.assertEqual(1, len(response["statistics"][0]["MockCommunity"]))
        self.assertIn("245:on_old_introduction_response", response["statistics"][0]["MockCommunity"])
        self.assertDictEqual(expected_stats,
                             response["statistics"][0]["MockCommunity"]["245:on_old_introduction_response"])

    async def test_statistics_one_overlay_with_unknown(self):
        """
        Check if statistics are returned for one loaded overlay, with an unknown message.
        """
        self.mount_statistics(0)
        self.node(0).endpoint.add_sent_stat(self.overlay(0).get_prefix(), 245, 1337, 42.0)
        self.node(0).endpoint.add_sent_stat(self.overlay(0).get_prefix(), 69, 1492, 7.0)  # 69 does not exist!

        self.assertIsNone(self.overlay(0).decode_map[69])  # Test invariant, use a number that does not exist.

        response = await self.make_request(self.ipv8, "overlays/statistics", "GET")

        self.assertIn("statistics", response)
        self.assertEqual(1, len(response["statistics"]))
        self.assertIn("MockCommunity", response["statistics"][0])
        self.assertEqual(2, len(response["statistics"][0]["MockCommunity"]))
        self.assertIn("245:on_old_introduction_response", response["statistics"][0]["MockCommunity"])
        self.assertIn("69:unknown", response["statistics"][0]["MockCommunity"])

    async def test_enable_stats_not_supported(self):
        """
        Check if stats cannot be enabled on an endpoint that is not a StatisticsEndpoint.
        """
        response = await self.make_request(self.ipv8, "overlays/statistics", "POST", json={'enable': 'true'},
                                           expected_status=412)

        self.assertFalse(response["success"])

    async def test_enable_stats_no_enable_param(self):
        """
        Check if stats cannot be enabled when the "enable" parameter is missing.
        """
        self.mount_statistics(0)

        response = await self.make_request(self.ipv8, "overlays/statistics", "POST", json={},
                                           expected_status=400)

        self.assertFalse(response["success"])

    async def test_enable_stats_no_target(self):
        """
        Check if stats cannot be enabled without specifying what overlay(s) to use.
        """
        self.mount_statistics(0)

        response = await self.make_request(self.ipv8, "overlays/statistics", "POST", json={'enable': 'true'},
                                           expected_status=412)

        self.assertFalse(response["success"])

    async def test_enable_stats_all(self):
        """
        Check if stats are correctly returned for one "all" overlays.
        """
        self.mount_statistics(0)

        response = await self.make_request(self.ipv8, "overlays/statistics", "POST",
                                           json={'enable': 'true', 'all': 'true'})

        self.assertTrue(response["success"])
        self.assertIn(self.overlay(0).get_prefix(), self.overlay(0).endpoint.statistics)

    async def test_enable_stats_all_many(self):
        """
        Check if stats are correctly returned for all overlays.
        """
        self.mount_statistics(0)
        mock_community2 = self.add_mock_community(0, MockCommunity2)

        response = await self.make_request(self.ipv8, "overlays/statistics", "POST",
                                           json={'enable': 'true', 'all': 'true'})

        self.assertTrue(response["success"])
        self.assertIn(self.overlay(0).get_prefix(), self.overlay(0).endpoint.statistics)
        self.assertIn(mock_community2.get_prefix(), self.overlay(0).endpoint.statistics)

    async def test_enable_stats_one_exclude(self):
        """
        Check if stats are correctly returned for a specific overlay, excluding another.
        """
        self.mount_statistics(0)
        mock_community2 = self.add_mock_community(0, MockCommunity2)

        response = await self.make_request(self.ipv8, "overlays/statistics", "POST",
                                           json={'enable': 'true', 'overlay_name': 'MockCommunity'})

        self.assertTrue(response["success"])
        self.assertIn(self.overlay(0).get_prefix(), self.overlay(0).endpoint.statistics)
        self.assertNotIn(mock_community2.get_prefix(), self.overlay(0).endpoint.statistics)

    async def test_enable_stats_one_include(self):
        """
        Check if stats are correctly returned for a specific overlay, including another.
        """
        self.mount_statistics(0)
        mock_community2 = self.add_mock_community(0, MockCommunity2)

        response = await self.make_request(self.ipv8, "overlays/statistics", "POST",
                                           json={'enable': 'true', 'overlay_name': 'MockCommunity'})
        response2 = await self.make_request(self.ipv8, "overlays/statistics", "POST",
                                            json={'enable': 'true', 'overlay_name': 'MockCommunity2'})

        self.assertTrue(response["success"])
        self.assertTrue(response2["success"])
        self.assertIn(self.overlay(0).get_prefix(), self.overlay(0).endpoint.statistics)
        self.assertIn(mock_community2.get_prefix(), self.overlay(0).endpoint.statistics)



from ..base import TestBase
from ..mocking.community import MockCommunity
from ...peerdiscovery.discovery import RandomWalk


class TestRandomWalk(TestBase):

    def setUp(self):
        super(TestRandomWalk, self).setUp()

        node_count = 3
        self.overlays = [MockCommunity() for _ in range(node_count)]
        self.strategies = [RandomWalk(self.overlays[i], reset_chance=0) for i in range(node_count)]

    async def tearDown(self):
        for overlay in self.overlays:
            await overlay.unload()
        return await super(TestRandomWalk, self).tearDown()

    async def test_take_step(self):
        """
        Check if we will walk to a random other node.

        Unit test network layout:
          NODE0 <-> NODE1 <-> NODE2
        """
        self.overlays[0].network.add_verified_peer(self.overlays[1].my_peer)
        self.overlays[0].network.discover_services(self.overlays[1].my_peer, [self.overlays[1].community_id, ])
        self.overlays[1].network.add_verified_peer(self.overlays[2].my_peer)
        self.overlays[1].network.discover_services(self.overlays[2].my_peer, [self.overlays[2].community_id, ])
        # We expect NODE1 to introduce NODE0 to NODE2
        self.strategies[0].take_step()
        await self.deliver_messages()
        self.strategies[0].take_step()
        await self.deliver_messages()

        self.assertEqual(len(self.overlays[0].network.verified_peers), 2)

    async def test_take_step_into(self):
        """
        Check if we will walk to an introduced node.

        Unit test network layout:
          NODE0 <-> (NODE1) <-> NODE2
          NODE0 -> NODE2
        """
        self.overlays[0].network.add_verified_peer(self.overlays[1].my_peer)
        self.overlays[0].network.discover_address(self.overlays[1].my_peer, self.overlays[2].endpoint.wan_address,
                                                  MockCommunity.community_id)
        self.overlays[0].network.discover_services(self.overlays[1].my_peer, [self.overlays[1].community_id, ])
        # We expect NODE0 to visit NODE2
        self.strategies[0].take_step()
        await self.deliver_messages()
        self.strategies[0].take_step()
        await self.deliver_messages()

        self.assertEqual(len(self.overlays[0].network.verified_peers), 2)

    async def test_fail_step_into(self):
        """
        Check if we drop an unreachable introduced node.

        Unit test network layout:
          NODE0 <-> (NODE1) <-> NODE2
          NODE0 -> NODE2
        """
        self.overlays[0].network.add_verified_peer(self.overlays[1].my_peer)
        self.overlays[0].network.discover_address(self.overlays[1].my_peer, self.overlays[2].endpoint.wan_address,
                                                  MockCommunity.community_id)
        self.overlays[0].network.discover_services(self.overlays[1].my_peer, [self.overlays[1].community_id, ])
        # Fail immediately when unreachable
        self.strategies[0].node_timeout = 0.0

        # NODE0 attempts to reach NODE2
        self.overlays[2].endpoint.close()
        self.strategies[0].take_step()
        # At this point the unreachable node should not have been removed yet
        self.assertEqual(len(self.overlays[0].network.get_walkable_addresses()), 1)

        await self.deliver_messages()

        # We expect NODE0 to clean unreachable NODE2
        self.strategies[0].take_step()

        self.assertEqual(len(self.overlays[0].network.get_walkable_addresses()), 0)
        self.assertEqual(len(self.overlays[0].network.verified_peers), 1)

    async def test_retry_step_into(self):
        """
        Check if we don't drop an introduced node immediately.

        Unit test network layout:
          NODE0 <-> (NODE1) <-> NODE2
          NODE0 -> NODE2
        """
        self.overlays[0].network.add_verified_peer(self.overlays[1].my_peer)
        self.overlays[0].network.discover_address(self.overlays[1].my_peer, self.overlays[2].endpoint.wan_address,
                                                  MockCommunity.community_id)
        self.overlays[0].network.discover_services(self.overlays[1].my_peer, [self.overlays[1].community_id, ])
        self.strategies[0].node_timeout = 100000.0

        # NODE0 attempts to reach NODE2
        self.overlays[2].endpoint.close()
        self.strategies[0].take_step()

        await self.deliver_messages()

        # NODE2 is still within its timeout and should not have been cleaned yet
        self.strategies[0].take_step()

        self.assertEqual(len(self.overlays[0].network.get_walkable_addresses()), 1)
        self.assertEqual(len(self.overlays[0].network.verified_peers), 1)



from binascii import unhexlify
from random import sample
from time import time

from .churn import DiscoveryStrategy, RandomChurn
from .payload import (DiscoveryIntroductionRequestPayload, PingPayload, PongPayload, SimilarityRequestPayload,
                      SimilarityResponsePayload)
from ..community import Community, DEFAULT_MAX_PEERS
from ..keyvault.crypto import default_eccrypto
from ..lazy_community import PacketDecodingError, lazy_wrapper, lazy_wrapper_unsigned, retrieve_cache
from ..messaging.payload import IntroductionRequestPayload
from ..messaging.payload_headers import BinMemberAuthenticationPayload, GlobalTimeDistributionPayload
from ..messaging.serialization import PackError
from ..peer import Peer
from ..requestcache import NumberCache, RequestCache


class PeriodicSimilarity(DiscoveryStrategy):

    def __init__(self, overlay):
        super(PeriodicSimilarity, self).__init__(overlay)
        self.last_step = 0

    def take_step(self):
        now = time()
        if (now - self.last_step < 1.0) or not self.overlay.network.verified_peers:
            return
        self.last_step = now
        with self.walk_lock:
            self.overlay.send_similarity_request(sample(self.overlay.network.verified_peers, 1)[0].address)


class PingRequestCache(NumberCache):

    name = "discoverypingcache"

    def __init__(self, request_cache, identifier, peer, start_time):
        super().__init__(request_cache, PingRequestCache.name, identifier)
        self.peer = peer
        self.start_time = start_time

    def finish(self):
        self.peer.pings.append(time() - self.start_time)

    @property
    def timeout_delay(self):
        return 5.0

    def on_timeout(self):
        pass


class DiscoveryCommunity(Community):

    version = b'\x02'
    community_id = unhexlify('7e313685c1912a141279f8248fc8db5899c5df5a')

    def __init__(self, my_peer, endpoint, network, max_peers=DEFAULT_MAX_PEERS, anonymize=False):
        super(DiscoveryCommunity, self).__init__(my_peer, endpoint, network, max_peers=max_peers, anonymize=anonymize)

        self.request_cache = RequestCache()

        self.add_message_handler(SimilarityRequestPayload, self.on_similarity_request)
        self.add_message_handler(SimilarityResponsePayload, self.on_similarity_response)
        self.add_message_handler(PingPayload, self.on_ping)
        self.add_message_handler(PongPayload, self.on_pong)

    def get_available_strategies(self):
        return {'PeriodicSimilarity': PeriodicSimilarity, 'RandomChurn': RandomChurn}

    async def unload(self):
        await self.request_cache.shutdown()
        await super(DiscoveryCommunity, self).unload()

    def on_old_introduction_request(self, source_address, data):
        if self.max_peers >= 0 and len(self.get_peers()) > self.max_peers:
            self.logger.debug("Dropping introduction request from (%s, %d): too many peers!",
                              source_address[0], source_address[1])
            return

        try:
            auth, dist, payload = self._ez_unpack_auth(DiscoveryIntroductionRequestPayload, data)
        except (PacketDecodingError, PackError):
            auth, dist, payload = self._ez_unpack_auth(IntroductionRequestPayload, data)

        peer = Peer(auth.public_key_bin, source_address)
        self.network.add_verified_peer(peer)
        self.network.discover_services(peer, [self.community_id, ])

        introduce_to = getattr(payload, 'introduce_to', None)
        introduction = None
        if introduce_to:
            peers = self.network.verified_peers
            matches = [p for p in peers if p.mid == introduce_to]
            introduction = matches[0] if matches else None
        packet = self.create_introduction_response(payload.destination_address, source_address, payload.identifier,
                                                   introduction=introduction, new_style=False)
        self.endpoint.send(source_address, packet)

    def introduction_response_callback(self, peer, dist, payload):
        self.send_similarity_request(peer.address)

    def send_similarity_request(self, address):
        my_peer_set = set([overlay.my_peer for overlay in self.network.service_overlays.values()])
        for peer in my_peer_set:
            packet = self.create_similarity_request(peer)
            self.endpoint.send(address, packet)

    @lazy_wrapper(GlobalTimeDistributionPayload, SimilarityRequestPayload)
    def on_similarity_request(self, node, dist, payload):
        self.network.discover_services(node, payload.preference_list)

        my_peer_set = set([overlay.my_peer for overlay in self.network.service_overlays.values()])
        for peer in my_peer_set:
            packet = self.create_similarity_response(payload.identifier, peer)
            self.endpoint.send(node.address, packet)

    @lazy_wrapper(GlobalTimeDistributionPayload, SimilarityResponsePayload)
    def on_similarity_response(self, node, dist, payload):
        if self.max_peers >= 0 and len(self.get_peers()) > self.max_peers and node not in self.network.verified_peers:
            self.logger.debug("Dropping similarity response from (%s, %d): too many peers!",
                              node.address[0], node.address[1])
            return

        self.network.add_verified_peer(node)
        self.network.discover_services(node, payload.preference_list)

    @lazy_wrapper_unsigned(GlobalTimeDistributionPayload, PingPayload)
    def on_ping(self, source_address, dist, payload):
        packet = self.create_pong(payload.identifier)
        self.endpoint.send(source_address, packet)

    @lazy_wrapper_unsigned(GlobalTimeDistributionPayload, PongPayload)
    @retrieve_cache(PingRequestCache)
    def on_pong(self, source_address, dist, payload, cache):
        cache.finish()

    def get_my_overlays(self, peer):
        return [service_id for service_id, overlay in self.network.service_overlays.items()
                if overlay.my_peer == peer]

    def custom_pack(self, peer, msg_num, payloads):
        packet = self._prefix + bytes([msg_num])
        packet += self.serializer.pack_serializable_list(payloads)
        packet += default_eccrypto.create_signature(peer.key, packet)
        return packet

    def create_similarity_request(self, peer):
        global_time = self.claim_global_time()
        payload = SimilarityRequestPayload(global_time,
                                           self.my_estimated_lan,
                                           self.my_estimated_wan,
                                           u"unknown",
                                           self.get_my_overlays(peer))
        auth = BinMemberAuthenticationPayload(peer.public_key.key_to_bin())
        dist = GlobalTimeDistributionPayload(global_time)

        return self.custom_pack(peer, 1, [auth, dist, payload])

    def create_similarity_response(self, identifier, peer):
        global_time = self.claim_global_time()
        payload = SimilarityResponsePayload(identifier, self.get_my_overlays(peer), [])
        auth = BinMemberAuthenticationPayload(peer.public_key.key_to_bin())
        dist = GlobalTimeDistributionPayload(global_time)

        return self.custom_pack(peer, 2, [auth, dist, payload])

    def send_ping(self, peer):
        global_time = self.claim_global_time()
        payload = PingPayload(global_time)
        dist = GlobalTimeDistributionPayload(global_time)

        packet = self._ez_pack(self._prefix, 3, [dist, payload], False)
        self.request_cache.add(PingRequestCache(self.request_cache, global_time, peer, time()))
        self.endpoint.send(peer.address, packet)

    def create_pong(self, identifier):
        global_time = self.claim_global_time()
        payload = PongPayload(identifier)
        dist = GlobalTimeDistributionPayload(global_time)
        return self._ez_pack(self._prefix, 4, [dist, payload], False)
